# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Set of HTML automatic forms to create, delete, copy or edit a single entity
or a list of entities of the same type
"""


from copy import copy

from logilab.common.registry import yes

from cubicweb import _
from cubicweb.predicates import (
    one_line_rset,
    non_final_entity,
    specified_etype_implements,
    is_instance,
)
from cubicweb.uilib import eid_param

from cubicweb_web.view import EntityView
from cubicweb_web import stdmsgs, formwidgets as fw
from cubicweb_web.form import FormViewMixIn
from cubicweb_web.views import uicfg, forms

_pvdc = uicfg.primaryview_display_ctrl


class DeleteConfForm(forms.CompositeForm):
    __regid__ = "deleteconf"
    # XXX non_final_entity does not implement eclass_selector
    __select__ = is_instance("Any")

    domid = "deleteconf"
    copy_nav_params = True
    form_buttons = [
        fw.Button(stdmsgs.BUTTON_DELETE, cwaction="delete"),
        fw.Button(
            stdmsgs.BUTTON_CANCEL,
            {"class": fw.Button.css_class + " cwjs-edition-cancel"},
        ),
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        done = set()
        for entity in self.cw_rset.entities():
            if entity.eid in done:
                continue
            done.add(entity.eid)
            subform = self._cw.vreg["forms"].select(
                "base", self._cw, entity=entity, mainform=False
            )
            self.add_subform(subform)


class DeleteConfFormView(FormViewMixIn, EntityView):
    """form used to confirm deletion of some entities"""

    __regid__ = "deleteconf"
    title = _("delete")
    # don't use navigation, all entities asked to be deleted should be displayed
    # else we will only delete the displayed page
    paginable = False
    # show first level of composite relations in a treeview
    show_composite = False
    show_composite_skip_rtypes = set(
        "wf_info_for",
    )

    def _iter_composite_entities(self, entity, limit=None):
        eids = set()
        for rdef, role in sorted(
            entity.e_schema.composite_rdef_roles, key=lambda x: x[0].relation_type
        ):
            if rdef.relation_type in self.show_composite_skip_rtypes:
                continue
            for centity in entity.related(
                rdef.relation_type, role, limit=limit
            ).entities():
                if centity.eid not in eids:
                    eids.add(centity.eid)
                    yield centity

    def call(self, onsubmit=None):
        """ask for confirmation before real deletion"""
        req, w = self._cw, self.w
        _ = req._
        if self.show_composite:
            req.add_css(
                ("jquery-treeview/jquery.treeview.css", "cubicweb.treeview.css")
            )
        w(
            "<script type=\"text/javascript\">updateMessage('%s');</script>\n",
            _("this action is not reversible!"),
        )
        # XXX above message should have style of a warning
        w("<h4>%s</h4>\n", _("Do you want to delete the following element(s)?"))
        form = self._cw.vreg["forms"].select(
            self.__regid__, req, rset=self.cw_rset, onsubmit=onsubmit
        )
        w("<ul>\n")
        page_size = req.property_value("navigation.page-size")
        for entity in self.cw_rset.entities():
            # don't use outofcontext view or any other that may contain inline
            # edition form
            w(
                '<li><a href="%s">%s</a>',
                entity.absolute_url(),
                entity.view("textoutofcontext"),
            )
            if self.show_composite:
                content = None
                for count, centity in enumerate(
                    self._iter_composite_entities(
                        entity,
                        limit=page_size,
                    )
                ):
                    if count == 0:
                        w('<ul class="treeview">')
                    if content is not None:
                        w('<li><a href="%s">%s</a></li>', *content)
                    if count == page_size - 1:
                        w(
                            '<li class="last">%s</li></ul>',
                            _("And more composite entities"),
                        )
                        break
                    content = (centity.absolute_url(), centity.view("textoutofcontext"))
                else:
                    if content is not None:
                        w('<li class="last"><a href="%s">%s</a></li></ul>', *content)
            w("</li>\n")
        w("</ul>\n")
        form.render(w=self.w)


class EditionFormView(FormViewMixIn, EntityView):
    """display primary entity edition form"""

    __regid__ = "edition"
    # add yes() so it takes precedence over deprecated views in baseforms,
    # though not baseforms based customized view
    __select__ = one_line_rset() & non_final_entity() & yes()
    form_id = "edition"

    title = _("modification")

    def cell_call(self, row, col, **kwargs):
        entity = self.cw_rset.complete_entity(row, col)
        self.render_form(entity)

    def render_form(self, entity):
        """fetch and render the form"""
        self.form_title(entity)
        form = self._cw.vreg["forms"].select(
            self.form_id, self._cw, entity=entity, submitmsg=self.submited_message()
        )
        self.init_form(form, entity)
        form.render(w=self.w)

    def init_form(self, form, entity):
        """customize your form before rendering here"""

    def form_title(self, entity):
        """the form view title"""
        ptitle = self._cw._(self.title)
        self.w(
            '<div class="formTitle"><span>%s %s</span></div>',
            entity.dc_type(),
            ptitle and "(%s)" % ptitle,
        )

    def submited_message(self):
        """return the message that will be displayed on successful edition"""
        return self._cw._("entity edited")


class CreationFormView(EditionFormView):
    """display primary entity creation form"""

    __regid__ = "creation"
    __select__ = specified_etype_implements("Any") & yes()

    title = _("creation")

    def call(self, **kwargs):
        """creation view for an entity"""
        # at this point we know etype is a valid entity type, thanks to our
        # selector
        etype = kwargs.pop("etype", self._cw.form.get("etype"))
        entity = self._cw.vreg["etypes"].etype_class(etype)(self._cw)
        entity.eid = next(self._cw.varmaker)
        self.render_form(entity)

    def form_title(self, entity):
        """the form view title"""
        if "__linkto" in self._cw.form:
            if isinstance(self._cw.form["__linkto"], list):
                # XXX which one should be considered (case: add a ticket to a
                # version in jpl)
                rtype, linkto_eid, role = self._cw.form["__linkto"][0].split(":")
            else:
                rtype, linkto_eid, role = self._cw.form["__linkto"].split(":")
            linkto_rset = self._cw.eid_rset(linkto_eid)
            linkto_type = linkto_rset.description[0][0]
            if role == "subject":
                title = self._cw.__(
                    "creating %s (%s %s %s %%(linkto)s)"
                    % (entity.e_schema, entity.e_schema, rtype, linkto_type)
                )
            else:
                title = self._cw.__(
                    "creating %s (%s %%(linkto)s %s %s)"
                    % (entity.e_schema, linkto_type, rtype, entity.e_schema)
                )
            msg = title % {"linkto": self._cw.view("incontext", linkto_rset)}
            self.w(
                '<div class="formTitle notransform"><span>%s</span></div>',
                msg,
                escape=False,
            )
        else:
            super().form_title(entity)

    def url(self):
        """return the url associated with this view"""
        req = self._cw
        return req.vreg["etypes"].etype_class(req.form["etype"]).cw_create_url(req)

    def submited_message(self):
        """return the message that will be displayed on successful edition"""
        return self._cw._("entity created")


class CopyFormView(EditionFormView):
    """display primary entity creation form initialized with values from another
    entity
    """

    __regid__ = "copy"

    title = _("copy")
    warning_message = _("Please note that this is only a shallow copy")

    def render_form(self, entity):
        """fetch and render the form"""
        # make a copy of entity to avoid altering the entity in the
        # request's cache.
        entity.complete()
        self.newentity = copy(entity)
        self.copying = entity
        self.newentity.eid = next(self._cw.varmaker)
        self.w(
            '<script type="text/javascript">updateMessage("%s");</script>\n',
            self._cw._(self.warning_message),
        )
        super().render_form(self.newentity)
        del self.newentity

    def init_form(self, form, entity):
        """customize your form before rendering here"""
        super().init_form(form, entity)
        if entity.eid == self.newentity.eid:
            form.add_hidden(eid_param("__cloned_eid", entity.eid), self.copying.eid)
        for rschema, role in form.editable_attributes():
            if not rschema.final:
                # ensure relation cache is filed
                rset = self.copying.related(rschema, role)
                self.newentity.cw_set_relation_cache(rschema, role, rset)

    def submited_message(self):
        """return the message that will be displayed on successful edition"""
        return self._cw._("entity copied")


class TableEditForm(forms.CompositeForm):
    __regid__ = "muledit"
    domid = "entityForm"
    onsubmit = "return validateForm('%s', null);" % domid
    form_buttons = [
        fw.SubmitButton(_("validate modifications on selected items")),
        fw.ResetButton(_("revert changes")),
    ]

    def __init__(self, req, rset, **kwargs):
        kwargs.setdefault("__redirectrql", rset.printable_rql())
        super().__init__(req, rset=rset, **kwargs)
        for row in range(len(self.cw_rset)):
            form = self._cw.vreg["forms"].select(
                "edition",
                self._cw,
                rset=self.cw_rset,
                row=row,
                formtype="muledit",
                copy_nav_params=False,
                mainform=False,
            )
            # XXX rely on the EntityCompositeFormRenderer to put the eid input
            form.remove_field(form.field_by_name("eid"))
            self.add_subform(form)


class TableEditFormView(FormViewMixIn, EntityView):
    __regid__ = "muledit"
    __select__ = EntityView.__select__ & yes()
    title = _("multiple edit")

    def call(self, **kwargs):
        """a view to edit multiple entities of the same type the first column
        should be the eid
        """
        # XXX overriding formvid (eg __form_id) necessary to make work edition:
        # the edit controller try to select the form with no rset but
        # entity=entity, and use this form to edit the entity. So we want
        # edition form there but specifying formvid may have other undesired
        # side effect. Maybe we should provide another variable optionally
        # telling which form the edit controller should select (eg difffers
        # between html generation / post handling form)
        form = self._cw.vreg["forms"].select(
            self.__regid__,
            self._cw,
            rset=self.cw_rset,
            copy_nav_params=True,
            formvid="edition",
        )
        form.render(w=self.w)



# pylint: disable=W0401,W0614
# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""exceptions used in the core of the CubicWeb web application"""

import http.client as http_client

from cubicweb import CubicWebException, Unauthorized  # noqa # backward compatibility
from cubicweb.utils import json_dumps


class DirectResponse(Exception):
    """Used to supply a twitted HTTP Response directly"""

    def __init__(self, response):
        self.response = response


class InvalidSession(CubicWebException):
    """raised when a session id is found but associated session is not found or
    invalid"""


# Publish related exception


class PublishException(CubicWebException):
    """base class for publishing related exception"""

    def __init__(self, *args, **kwargs):
        self.status = kwargs.pop("status", http_client.OK)
        super().__init__(*args, **kwargs)


class Redirect(PublishException):
    """raised to redirect the http request"""

    def __init__(self, location, status=http_client.SEE_OTHER):
        super().__init__(status=status)
        self.location = location


class LogOut(PublishException):
    """raised to ask for deauthentication of a logged in user"""

    def __init__(self, url=None):
        super().__init__()
        self.url = url


class RequestError(PublishException):
    """raised when a request can't be served because of a bad input"""

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("status", http_client.BAD_REQUEST)
        super().__init__(*args, **kwargs)


class RemoteCallFailed(RequestError):
    """raised when a json remote call fails"""

    def __init__(self, reason="", status=http_client.INTERNAL_SERVER_ERROR):
        super().__init__(reason, status=status)
        self.reason = reason

    def dumps(self):
        return json_dumps({"reason": self.reason})


class NotFound(RequestError):
    """raised when something was not found. In most case,
    a 404 error should be returned"""

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("status", http_client.NOT_FOUND)
        super().__init__(*args, **kwargs)


class StatusResponse(PublishException):
    def __init__(self, status, content=""):
        super().__init__(status=status)
        self.content = content

    def __repr__(self):
        return f"{self.__class__.__name__}({self.status!r}, {self.content!r})"


# Publish related error


class NothingToEdit(RequestError):
    """raised when an edit request doesn't specify any eid to edit"""

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("status", http_client.BAD_REQUEST)
        super().__init__(*args, **kwargs)


class ProcessFormError(RequestError):
    """raised when posted data can't be processed by the corresponding field"""

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("status", http_client.BAD_REQUEST)
        super().__init__(*args, **kwargs)



# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""cubicweb.web.controller unit tests

"""

from datetime import datetime, date, time

from logilab.common.testlib import unittest_main

from cubicweb_web.devtools.testlib import WebCWTC


class BaseControllerCWTC(WebCWTC):
    def test_parse_datetime_ok(self):
        with self.admin_access.web_request() as req:
            ctrl = self.vreg["controllers"].select("view", req)
            pd = ctrl._cw.parse_datetime
            self.assertIsInstance(pd("2006/06/24 12:18"), datetime)
            self.assertIsInstance(pd("2006/06/24"), date)
            self.assertIsInstance(pd("2006/06/24 12:18", "Datetime"), datetime)
            self.assertIsInstance(pd("2006/06/24", "Datetime"), datetime)
            self.assertIsInstance(pd("2006/06/24", "Date"), date)
            self.assertIsInstance(pd("12:18", "Time"), time)

    def test_parse_datetime_ko(self):
        with self.admin_access.web_request() as req:
            ctrl = self.vreg["controllers"].select("view", req)
            pd = ctrl._cw.parse_datetime
            self.assertRaises(ValueError, pd, "2006/06/24 12:188", "Datetime")
            self.assertRaises(ValueError, pd, "2006/06/240", "Datetime")
            self.assertRaises(ValueError, pd, "2006/06/24 12:18", "Date")
            self.assertRaises(ValueError, pd, "2006/24/06", "Date")
            self.assertRaises(ValueError, pd, "2006/06/240", "Date")
            self.assertRaises(ValueError, pd, "12:188", "Time")


if __name__ == "__main__":
    unittest_main()



import os

from cubicweb.cwvreg import CWRegistryStore
from cubicweb.i18n import add_msg
from cubicweb.devtools.devctl import (
    I18nCubeMessageExtractor,
    _iter_vreg_objids,
    DevConfiguration,
    clean_sys_modules,
)

from cubicweb_web.webconfig import WebAllInOneConfiguration


class WebI18nCubeMessageExtractor(I18nCubeMessageExtractor):
    def schemapot(self) -> str:
        """append uicfg related pot to the generated ``schema.pot``"""
        schemapot_path = super().schemapot()
        print("schema(cubicweb_web)", end=" ")

        # append to the existing pot file
        schemapotstream = open(schemapot_path, "a")
        generate_schema_pot(schemapotstream.write, self.cubedir)
        schemapotstream.close()

        return schemapot_path


class WebDevConfiguration(WebAllInOneConfiguration, DevConfiguration):
    pass


def generate_schema_pot(w, cubedir=None):
    """generate a pot file with schema specific i18n messages
    notice that relation definitions description and static vocabulary
    should be marked using '_' and extracted using xgettext
    """

    cube = os.path.split(cubedir)[-1]
    if cube.startswith("cubicweb_"):
        cube = cube[len("cubicweb_") :]

    # XXX Cube `web` must be installed in order for i18ncube to work as we
    # use appobjects registered by it in _generate_schema_pot.
    config = WebDevConfiguration("web", cube)
    depcubes = list(config._cubes)
    depcubes.remove(cube)
    if "web" not in depcubes:
        depcubes.insert(0, "web")
    libconfig = WebDevConfiguration(*depcubes)

    clean_sys_modules(config.appobjects_modnames())
    schema = config.load_schema(remove_unused_relation_types=False)
    vreg = CWRegistryStore(config)
    # set_schema triggers objects registrations
    vreg.set_schema(schema)
    _generate_schema_pot(w, vreg, schema, libconfig=libconfig)


def _generate_schema_pot(w, vreg, schema, libconfig=None):
    afss = vreg["uicfg"]["autoform_section"]
    aiams = vreg["uicfg"]["actionbox_appearsin_addmenu"]

    libschema = libconfig.load_schema(remove_unused_relation_types=False)
    clean_sys_modules(libconfig.appobjects_modnames())
    libvreg = CWRegistryStore(libconfig)
    libvreg.set_schema(libschema)  # trigger objects registration

    libafss = libvreg["uicfg"]["autoform_section"]
    libaiams = libvreg["uicfg"]["actionbox_appearsin_addmenu"]

    for eschema in sorted(schema.entities()):
        if eschema.final:
            continue

        for rschema, targetschemas, role in eschema.relation_definitions(True):
            if rschema.final:
                continue
            for tschema in targetschemas:
                for afs in afss:
                    fsections = afs.etype_get(eschema, rschema, role, tschema)
                    if "main_inlined" in fsections and not _is_in_lib(
                        libafss,
                        eschema,
                        rschema,
                        role,
                        tschema,
                        lambda x: "main_inlined" in x,
                    ):
                        add_msg(
                            w,
                            "add a %s" % tschema,
                            "inlined:%s.%s.%s" % (eschema.type, rschema, role),
                        )
                        add_msg(
                            w,
                            str(tschema),
                            "inlined:%s.%s.%s" % (eschema.type, rschema, role),
                        )
                        break
                for aiam in aiams:
                    if aiam.etype_get(
                        eschema, rschema, role, tschema
                    ) and not _is_in_lib(libaiams, eschema, rschema, role, tschema):
                        if role == "subject":
                            label = "add %s %s %s %s" % (
                                eschema,
                                rschema,
                                tschema,
                                role,
                            )
                            label2 = "creating %s (%s %%(linkto)s %s %s)" % (
                                tschema,
                                eschema,
                                rschema,
                                tschema,
                            )
                        else:
                            label = "add %s %s %s %s" % (
                                tschema,
                                rschema,
                                eschema,
                                role,
                            )
                            label2 = "creating %s (%s %s %s %%(linkto)s)" % (
                                tschema,
                                tschema,
                                rschema,
                                eschema,
                            )
                        add_msg(w, label)
                        add_msg(w, label2)
                        break
    vregdone = set()
    # fill vregdone with already processed registry
    list(_iter_vreg_objids(libvreg, vregdone))
    for objid in _iter_vreg_objids(vreg, vregdone):
        add_msg(w, f"{objid}_description")
        add_msg(w, objid)


def _is_in_lib(rtags, eschema, rschema, role, tschema, predicate=bool):
    return any(
        predicate(rtag.etype_get(eschema, rschema, role, tschema)) for rtag in rtags
    )



# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

import collections
import hashlib
import tempfile
from json import loads
from os.path import join
from urllib.parse import urljoin

from cubicweb.devtools import BASE_URL
from logilab.common.testlib import TestCase, unittest_main
from webtest.forms import Upload

from cubicweb_web.devtools.testlib import FakeRequest, PyramidWebCWTC


class AjaxReplaceUrlTC(TestCase):
    def test_ajax_replace_url_1(self):
        self._test_arurl(
            "fname=view&rql=Person%20P&vid=list", rql="Person P", vid="list"
        )

    def test_ajax_replace_url_2(self):
        self._test_arurl(
            "age=12&fname=view&name=bar&rql=Person%20P&vid=oneline",
            rql="Person P",
            vid="oneline",
            name="bar",
            age=12,
        )

    def _test_arurl(self, qs, **kwargs):
        req = FakeRequest()
        arurl = req.ajax_replace_url
        # NOTE: for the simplest use cases, we could use doctest
        url = arurl("foo", **kwargs)
        self.assertTrue(url.startswith("javascript:"))
        self.assertTrue(url.endswith("()"))
        cbname = url.split()[1][:-2]
        self.assertMultiLineEqual(
            f'function {cbname}() {{ $("#foo").loadxhtml("{BASE_URL}ajax?{qs}",'
            f'{{"pageid": "{req.pageid}"}},"get","replace"); }}',
            req.html_headers.post_inlined_scripts[0],
        )


class FileUploadTC(PyramidWebCWTC):
    def _fobject(self, fname):
        return open(join(self.datadir, fname), "rb")

    def _fcontent(self, fname):
        with self._fobject(fname) as f:
            return f.read()

    def _fhash(self, fname):
        content = self._fcontent(fname)
        return hashlib.md5(content).hexdigest()

    def test_single_file_upload(self):
        webreq = self.webapp.post(
            urljoin(BASE_URL, "ajax?fname=fileupload"),
            collections.OrderedDict(
                [("file", Upload("schema.py", self._fobject("schema.py").read()))]
            ),
        )

        # check backward compat : a single uploaded file leads to a single
        # 2-uple in the request form
        expect = {
            "fname": "fileupload",
            "file": ["schema.py", self._fhash("schema.py")],
        }

        body = loads(webreq.body)
        del body["csrf_token"]

        self.assertEqual(webreq.status_code, 200)
        self.assertDictEqual(expect, body)

    def test_multiple_file_upload(self):
        files = [
            ("files", Upload("schema.py", self._fobject("schema.py").read())),
            ("files", Upload("views.py", self._fobject("views.py").read())),
        ]

        webreq = self.webapp.post(
            urljoin(BASE_URL, "ajax?fname=fileupload"),
            files,
            headers={"X-CSRF-Token": self.webapp.get_csrf_token()},
        )

        expect = {
            "fname": "fileupload",
            "files": [
                ["schema.py", self._fhash("schema.py")],
                ["views.py", self._fhash("views.py")],
            ],
        }
        self.assertEqual(webreq.status_code, 200)
        self.assertDictEqual(expect, loads(webreq.text))


class LanguageTC(PyramidWebCWTC):
    def test_language_neg(self):
        headers = {"Accept-Language": "fr"}
        webreq = self.webapp.get(BASE_URL, headers=headers)
        webreq.mustcontain(b'lang="fr"')
        vary = [h.lower().strip() for h in webreq.headers.get("Vary").split(",")]
        self.assertIn("accept-language", vary)
        headers = {"Accept-Language": "en"}
        webreq = self.webapp.get(BASE_URL, headers=headers)
        webreq.mustcontain(b'lang="en"')
        vary = [h.lower().strip() for h in webreq.headers.get("Vary").split(",")]
        self.assertIn("accept-language", vary)

    def test_response_codes(self):
        with self.admin_access.client_cnx() as cnx:
            admin_eid = cnx.user.eid
        # guest can't see admin
        request = self.webapp.get(urljoin(BASE_URL, str(admin_eid)), status=403)
        self.assertEqual(request.status_code, 403)

        # but admin can
        self.login()
        request = self.webapp.get(urljoin(BASE_URL, str(admin_eid)))
        self.assertEqual(request.status_code, 200)

    def test_session_cookie_httponly(self):
        webreq = self.webapp.get(BASE_URL)
        self.assertIn("HttpOnly", "".join(webreq.headers.getall("Set-Cookie")))


class MiscOptionsTC(PyramidWebCWTC):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.logfile = tempfile.NamedTemporaryFile()

    @classmethod
    def init_config(cls, config):
        super().init_config(config)
        config.global_set_option("query-log-file", cls.logfile.name)
        config.global_set_option("datadir-url", "//static.testing.cubicweb/")
        # call load_configuration again to let the config reset its datadir_url
        config.load_configuration()

    def test_log_queries(self):
        self.webapp.get(BASE_URL)
        self.assertTrue(self.logfile.read())

    def test_datadir_url(self):
        webreq = self.webapp.get(BASE_URL)
        self.assertNotIn(b"/data/", webreq.body)

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        cls.logfile.close()


if __name__ == "__main__":
    unittest_main()



# copyright 2014-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from logilab.common import tempattr, attrdict

from cubicweb_web.devtools.testlib import WebCWTC
from cubicweb_web.views.autoform import InlinedFormField
from cubicweb_web.views.forms import EntityFieldsForm


class InlinedFormCWTC(WebCWTC):
    def test_linked_to(self):
        with self.admin_access.web_request() as req:
            formview = req.vreg["views"].select(
                "inline-creation",
                req,
                etype="File",
                rtype="described_by_test",
                role="subject",
                peid=123,
                petype="Salesterm",
            )
            self.assertEqual(
                {("described_by_test", "object"): [123]}, formview.form.linked_to
            )

    def test_linked_to_parent_being_created(self):
        with self.admin_access.web_request() as req:
            formview = req.vreg["views"].select(
                "inline-creation",
                req,
                etype="File",
                rtype="described_by_test",
                role="subject",
                peid="A",
                petype="Salesterm",
            )
            self.assertEqual(formview.form.linked_to, {})

    def test_remove_js_depending_on_cardinality(self):
        with self.admin_access.web_request() as req:
            formview = req.vreg["views"].select(
                "inline-creation",
                req,
                etype="File",
                rtype="described_by_test",
                role="subject",
                peid="A",
                petype="Salesterm",
            )
            # cardinality is 1, can't remove
            self.assertIsNone(formview._get_removejs())
            rdef = self.schema["Salesterm"].relation_definition("described_by_test")
            with tempattr(rdef, "cardinality", "?*"):
                self.assertTrue(formview._get_removejs())
            with tempattr(rdef, "cardinality", "+*"):
                # formview has no parent info (pform). This is what happens
                # when an inline form is requested through AJAX.
                self.assertTrue(formview._get_removejs())
                fakeview = attrdict(dict(rtype="described_by_test", role="subject"))
                # formview is first, can't be removed
                formview.pform = attrdict(
                    fields=[
                        InlinedFormField(view=formview),
                        InlinedFormField(view=fakeview),
                    ]
                )
                self.assertIsNone(formview._get_removejs())
                # formview isn't first, can be removed
                formview.pform = attrdict(
                    fields=[
                        InlinedFormField(view=fakeview),
                        InlinedFormField(view=formview),
                    ]
                )
                self.assertTrue(formview._get_removejs())

    def test_field_by_name_consider_aff(self):
        class MyField:
            def __init__(self, *args, **kwargs):
                pass

        EntityFieldsForm.uicfg_aff.tag_attribute(("CWUser", "firstname"), MyField)
        try:
            with self.admin_access.web_request() as req:
                form = req.vreg["forms"].select("base", req, entity=req.user)
                self.assertIsInstance(
                    form.field_by_name("firstname", "subject", req.user.e_schema),
                    MyField,
                )
        finally:
            EntityFieldsForm.uicfg_aff.del_rtag("CWUser", "firstname", "*", "subject")


if __name__ == "__main__":
    import unittest

    unittest.main()



# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""workflow views:

* IWorkflowable views and forms
* workflow entities views (State, Transition, TrInfo)
"""


from logilab.mtconverter import xml_escape

from cubicweb import Unauthorized
from cubicweb import _
from cubicweb.predicates import (
    one_line_rset,
    relation_possible,
    match_form_params,
    score_entity,
    is_instance,
    adaptable,
)
from cubicweb_web.view import EntityView
from cubicweb_web import formwidgets as fwdgs
from cubicweb_web import stdmsgs, action, component, form
from cubicweb_web.views import uicfg, forms, ibreadcrumbs
from cubicweb_web.views.dotgraphview import DotGraphView, DotPropsHandler
from cubicweb_web.views.tabs import TabbedPrimaryView, PrimaryTab

_pvs = uicfg.primaryview_section
_pvs.tag_subject_of(("Workflow", "initial_state", "*"), "hidden")
_pvs.tag_object_of(("*", "state_of", "Workflow"), "hidden")
_pvs.tag_object_of(("*", "transition_of", "Workflow"), "hidden")
_pvs.tag_object_of(("*", "wf_info_for", "*"), "hidden")
for rtype in ("in_state", "by_transition", "from_state", "to_state"):
    _pvs.tag_subject_of(("*", rtype, "*"), "hidden")
    _pvs.tag_object_of(("*", rtype, "*"), "hidden")
_pvs.tag_object_of(("*", "wf_info_for", "*"), "hidden")

_abaa = uicfg.actionbox_appearsin_addmenu
_abaa.tag_subject_of(("BaseTransition", "condition", "RQLExpression"), False)
_abaa.tag_subject_of(("State", "allowed_transition", "BaseTransition"), False)
_abaa.tag_object_of(("SubWorkflowExitPoint", "destination_state", "State"), False)
_abaa.tag_subject_of(("*", "wf_info_for", "*"), False)
_abaa.tag_object_of(("*", "wf_info_for", "*"), False)

_abaa.tag_object_of(("*", "state_of", "CWEType"), True)
_abaa.tag_object_of(("*", "transition_of", "CWEType"), True)
_abaa.tag_subject_of(("Transition", "destination_state", "*"), True)
_abaa.tag_object_of(("*", "allowed_transition", "Transition"), True)
_abaa.tag_object_of(("*", "destination_state", "State"), True)
_abaa.tag_subject_of(("State", "allowed_transition", "*"), True)
_abaa.tag_object_of(("State", "state_of", "Workflow"), True)
_abaa.tag_object_of(("Transition", "transition_of", "Workflow"), True)
_abaa.tag_object_of(("WorkflowTransition", "transition_of", "Workflow"), True)

_afs = uicfg.autoform_section
_affk = uicfg.autoform_field_kwargs


# IWorkflowable views #########################################################


class ChangeStateForm(forms.CompositeEntityForm):
    # set dom id to ensure there is no conflict with edition form (see
    # session_key() implementation)
    __regid__ = domid = "changestate"

    form_renderer_id = "base"  # don't want EntityFormRenderer
    form_buttons = [
        fwdgs.SubmitButton(),
        fwdgs.Button(
            stdmsgs.BUTTON_CANCEL,
            {"class": fwdgs.Button.css_class + " cwjs-edition-cancel"},
        ),
    ]


class ChangeStateFormView(form.FormViewMixIn, EntityView):
    __regid__ = "statuschange"
    title = _("status change")
    __select__ = (
        one_line_rset() & match_form_params("treid") & adaptable("IWorkflowable")
    )

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        transition = self._cw.entity_from_eid(self._cw.form["treid"])
        form = self.get_form(entity, transition)
        self.w(
            "<h4>%s %s</h4>\n",
            self._cw._(transition.name),
            entity.view("oneline"),
            escape=False,
        )
        msg = self._cw._("status will change from %(st1)s to %(st2)s") % {
            "st1": entity.cw_adapt_to("IWorkflowable").printable_state,
            "st2": self._cw._(transition.destination(entity).name),
        }
        self.w("<p>%s</p>\n", msg)
        form.render(w=self.w)

    def redirectpath(self, entity):
        return entity.rest_path()

    def get_form(self, entity, transition, **kwargs):
        # XXX used to specify both rset/row/col and entity in case implements
        # selector (and not is_instance) is used on custom form
        form = self._cw.vreg["forms"].select(
            "changestate",
            self._cw,
            entity=entity,
            transition=transition,
            redirect_path=self.redirectpath(entity),
            **kwargs,
        )
        trinfo = self._cw.vreg["etypes"].etype_class("TrInfo")(self._cw)
        trinfo.eid = next(self._cw.varmaker)
        subform = self._cw.vreg["forms"].select(
            "edition", self._cw, entity=trinfo, mainform=False
        )
        subform.field_by_name("wf_info_for", "subject").value = entity.eid
        trfield = subform.field_by_name("by_transition", "subject")
        trfield.widget = fwdgs.HiddenInput()
        trfield.value = transition.eid
        form.add_subform(subform)
        return form


class WFHistoryView(EntityView):
    __regid__ = "wfhistory"
    __select__ = relation_possible("wf_info_for", role="object") & score_entity(
        lambda x: x.cw_adapt_to("IWorkflowable").workflow_history
    )

    title = _("Workflow history")

    def cell_call(self, row, col, view=None, title=title):
        _ = self._cw._
        eid = self.cw_rset[row][col]
        sel = "Any FS,TS,C,D"
        rql = (
            " ORDERBY D DESC WHERE WF wf_info_for X,"
            "WF from_state FS, WF to_state TS, WF comment C,"
            "WF creation_date D"
        )
        if self._cw.vreg.schema.eschema("CWUser").has_perm(self._cw, "read"):
            sel += ",U,WF"
            rql += ", WF owned_by U?"
            headers = (
                _("from_state"),
                _("to_state"),
                _("comment"),
                _("date"),
                _("CWUser"),
            )
        else:
            sel += ",WF"
            headers = (_("from_state"), _("to_state"), _("comment"), _("date"))
        sel += ",FSN,TSN,CF"
        rql = "%s %s, FS name FSN, TS name TSN, WF comment_format CF, X eid %%(x)s" % (
            sel,
            rql,
        )
        try:
            rset = self._cw.execute(rql, {"x": eid})
        except Unauthorized:
            return
        if rset:
            if title:
                self.w("<h2>%s</h2>\n", _(title))
            self.wview("table", rset, headers=headers, cellvids={2: "editable-final"})


class WFHistoryVComponent(component.EntityCtxComponent):
    """display the workflow history for entities supporting it"""

    __regid__ = "wfhistory"
    __select__ = component.EntityCtxComponent.__select__ & WFHistoryView.__select__
    context = "navcontentbottom"
    title = _("Workflow history")

    def render_body(self, w):
        self.entity.view("wfhistory", w=w, title=None)


class InContextWithStateView(EntityView):
    """display incontext view for an entity as well as its current state"""

    __regid__ = "incontext-state"
    __select__ = adaptable("IWorkflowable")

    def entity_call(self, entity):
        iwf = entity.cw_adapt_to("IWorkflowable")
        self.w("%s [%s]", entity.view("incontext"), iwf.printable_state)


# workflow actions #############################################################


class WorkflowActions(action.Action):
    """fill 'workflow' sub-menu of the actions box"""

    __regid__ = "workflow"
    __select__ = (
        action.Action.__select__ & one_line_rset() & relation_possible("in_state")
    )

    submenu = _("workflow")
    order = 10

    def fill_menu(self, box, menu):
        entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        menu.label = "%s: %s" % (
            self._cw._("state"),
            entity.cw_adapt_to("IWorkflowable").printable_state,
        )
        menu.append_anyway = True
        super(WorkflowActions, self).fill_menu(box, menu)

    def actual_actions(self):
        entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        iworkflowable = entity.cw_adapt_to("IWorkflowable")
        hastr = False
        for tr in iworkflowable.possible_transitions():
            url = entity.absolute_url(vid="statuschange", treid=tr.eid)
            yield self.build_action(self._cw._(tr.name), url)
            hastr = True
        # don't propose to see wf if user can't pass any transition
        if hastr:
            wfurl = iworkflowable.current_workflow.absolute_url()
            yield self.build_action(self._cw._("view workflow"), wfurl)
        if iworkflowable.workflow_history:
            wfurl = entity.absolute_url(vid="wfhistory")
            yield self.build_action(self._cw._("view history"), wfurl)


# workflow entity types views ##################################################

_pvs = uicfg.primaryview_section
_pvs.tag_subject_of(("Workflow", "initial_state", "*"), "hidden")
_pvs.tag_object_of(("*", "state_of", "Workflow"), "hidden")
_pvs.tag_object_of(("*", "transition_of", "Workflow"), "hidden")
_pvs.tag_object_of(("*", "default_workflow", "Workflow"), "hidden")

_abaa = uicfg.actionbox_appearsin_addmenu
_abaa.tag_subject_of(("BaseTransition", "condition", "RQLExpression"), False)
_abaa.tag_subject_of(("State", "allowed_transition", "BaseTransition"), False)
_abaa.tag_object_of(("SubWorkflowExitPoint", "destination_state", "State"), False)
_abaa.tag_object_of(("State", "state_of", "Workflow"), True)
_abaa.tag_object_of(("BaseTransition", "transition_of", "Workflow"), False)
_abaa.tag_object_of(("Transition", "transition_of", "Workflow"), True)
_abaa.tag_object_of(("WorkflowTransition", "transition_of", "Workflow"), True)


class WorkflowPrimaryView(TabbedPrimaryView):
    __select__ = is_instance("Workflow")
    tabs = [_("wf_tab_info"), _("wfgraph")]
    default_tab = "wf_tab_info"


class StateInContextView(EntityView):
    """convenience trick, State's incontext view should not be clickable"""

    __regid__ = "incontext"
    __select__ = is_instance("State")

    def cell_call(self, row, col):
        self.w(
            xml_escape(self._cw.view("textincontext", self.cw_rset, row=row, col=col))
        )


class WorkflowTabTextView(PrimaryTab):
    __regid__ = "wf_tab_info"
    __select__ = PrimaryTab.__select__ & one_line_rset() & is_instance("Workflow")

    def render_entity_attributes(self, entity):
        _ = self._cw._
        self.w("<div>%s</div>", entity.printable_value("description"))
        self.w("<span>%s%s</span>", _("workflow_of").capitalize(), _(" :"))
        html = []
        for e in entity.workflow_of:
            view = e.view("outofcontext")
            if entity.eid == e.default_workflow[0].eid:
                view += " <span>[%s]</span>" % _("default_workflow")
            html.append(view)
        self.w(", ".join(v for v in html))
        self.w("<h2>%s</h2>", _("Transition_plural"))
        rset = self._cw.execute(
            "Any T,T,DS,T,TT ORDERBY TN WHERE T transition_of WF, WF eid %(x)s,"
            "T type TT, T name TN, T destination_state DS?",
            {"x": entity.eid},
        )
        self.wview(
            "table",
            rset,
            "null",
            cellvids={1: "trfromstates", 2: "outofcontext", 3: "trsecurity"},
            headers=(
                _("Transition"),
                _("from_state"),
                _("to_state"),
                _("permissions"),
                _("type"),
            ),
        )


class TransitionSecurityTextView(EntityView):
    __regid__ = "trsecurity"
    __select__ = is_instance("Transition")

    def cell_call(self, row, col):
        _ = self._cw._
        entity = self.cw_rset.get_entity(self.cw_row, self.cw_col)
        if entity.require_group:
            self.w(
                "<div>%s%s %s</div>",
                _("groups"),
                _(" :"),
                ", ".join(g.view("incontext") for g in entity.require_group),
            )
        if entity.condition:
            self.w(
                "<div>%s%s %s</div>",
                _("conditions"),
                _(" :"),
                "<br/>".join(e.dc_title() for e in entity.condition),
            )


class TransitionAllowedTextView(EntityView):
    __regid__ = "trfromstates"
    __select__ = is_instance("Transition")

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(self.cw_row, self.cw_col)
        self.w(
            ", ".join(e.view("outofcontext") for e in entity.reverse_allowed_transition)
        )


# workflow entity types edition ################################################


def _wf_items_for_relation(req, wfeid, wfrelation, field):
    wf = req.entity_from_eid(wfeid)
    rschema = req.vreg.schema[field.name]
    param = "toeid" if field.role == "subject" else "fromeid"
    return sorted(
        (e.view("combobox"), str(e.eid))
        for e in getattr(wf, "reverse_%s" % wfrelation)
        if rschema.has_perm(req, "add", **{param: e.eid})
    )


# TrInfo
_afs.tag_subject_of(("TrInfo", "to_state", "*"), "main", "hidden")
_afs.tag_subject_of(("TrInfo", "from_state", "*"), "main", "hidden")
_afs.tag_attribute(("TrInfo", "tr_count"), "main", "hidden")


# BaseTransition
# XXX * allowed_transition BaseTransition
# XXX BaseTransition destination_state *


def transition_states_vocabulary(form, field):
    entity = form.edited_entity
    if entity.has_eid():
        wfeid = entity.transition_of[0].eid
    else:
        eids = form.linked_to.get(("transition_of", "subject"))
        if not eids:
            return []
        wfeid = eids[0]
    return _wf_items_for_relation(form._cw, wfeid, "state_of", field)


_afs.tag_subject_of(("*", "destination_state", "*"), "main", "attributes")
_affk.tag_subject_of(
    ("*", "destination_state", "*"), {"choices": transition_states_vocabulary}
)
_afs.tag_object_of(("*", "allowed_transition", "*"), "main", "attributes")
_affk.tag_object_of(
    ("*", "allowed_transition", "*"), {"choices": transition_states_vocabulary}
)


# State


def state_transitions_vocabulary(form, field):
    entity = form.edited_entity
    if entity.has_eid():
        wfeid = entity.state_of[0].eid
    else:
        eids = form.linked_to.get(("state_of", "subject"))
        if not eids:
            return []
        wfeid = eids[0]
    return _wf_items_for_relation(form._cw, wfeid, "transition_of", field)


_afs.tag_subject_of(("State", "allowed_transition", "*"), "main", "attributes")
_affk.tag_subject_of(
    ("State", "allowed_transition", "*"), {"choices": state_transitions_vocabulary}
)


# adaptaters ###################################################################


class WorkflowIBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    __select__ = is_instance("Workflow")

    # XXX what if workflow of multiple types?
    def parent_entity(self):
        return self.entity.workflow_of and self.entity.workflow_of[0] or None


class WorkflowItemIBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    __select__ = is_instance("BaseTransition", "State")

    def parent_entity(self):
        return self.entity.workflow


class TransitionItemIBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    __select__ = is_instance("SubWorkflowExitPoint")

    def parent_entity(self):
        return self.entity.reverse_subworkflow_exit[0]


class TrInfoIBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    __select__ = is_instance("TrInfo")

    def parent_entity(self):
        return self.entity.for_entity


# workflow images ##############################################################


class WorkflowDotPropsHandler(DotPropsHandler):
    def node_properties(self, stateortransition):
        """return default DOT drawing options for a state or transition"""
        props = super(WorkflowDotPropsHandler, self).node_properties(stateortransition)
        if hasattr(stateortransition, "state_of"):
            props["shape"] = "box"
            props["style"] = "filled"
            if stateortransition.reverse_initial_state:
                props["fillcolor"] = "#88CC88"
        else:
            props["shape"] = "ellipse"
        return props


class WorkflowVisitor(object):
    def __init__(self, entity):
        self.entity = entity

    def nodes(self):
        for state in self.entity.reverse_state_of:
            state.complete()
            yield state.eid, state
        for transition in self.entity.reverse_transition_of:
            transition.complete()
            yield transition.eid, transition

    def edges(self):
        for transition in self.entity.reverse_transition_of:
            for incomingstate in transition.reverse_allowed_transition:
                yield incomingstate.eid, transition.eid, transition
            for outgoingstate in transition.potential_destinations():
                yield transition.eid, outgoingstate.eid, transition


class WorkflowGraphView(DotGraphView):
    __regid__ = "wfgraph"
    __select__ = EntityView.__select__ & one_line_rset() & is_instance("Workflow")

    def build_visitor(self, entity):
        return WorkflowVisitor(entity)

    def build_dotpropshandler(self):
        return WorkflowDotPropsHandler(self._cw)



"""misc. unittests for utility functions
"""

import unittest
from functools import partial
from urllib.parse import urljoin

from cubicweb import ObjectNotFound
from cubicweb.devtools import BASE_URL
from cubicweb.devtools.fake import FakeConfig, FakeCWRegistryStore

from cubicweb_web.devtools.testlib import WebCWTC
from cubicweb_web.request import (
    CubicWebRequestBase,
    _parse_accept_header,
    _mimetype_sort_key,
    _mimetype_parser,
    _charset_sort_key,
)


class AcceptParserTC(unittest.TestCase):
    def test_parse_accept(self):
        parse_accept_header = partial(
            _parse_accept_header,
            value_parser=_mimetype_parser,
            value_sort_key=_mimetype_sort_key,
        )
        # compare scores
        self.assertEqual(
            parse_accept_header("audio/*;q=0.2, audio/basic"),
            [
                (("audio/basic", ("audio", "basic", {}), 1.0)),
                (("audio/*", ("audio", "*", {}), 0.2)),
            ],
        )
        self.assertEqual(
            parse_accept_header(
                "text/plain;q=0.5, text/html, text/x-dvi;q=0.8, text/x-c"
            ),
            [
                (("text/html", ("text", "html", {}), 1.0)),
                (("text/x-c", ("text", "x-c", {}), 1.0)),
                (("text/x-dvi", ("text", "x-dvi", {}), 0.8)),
                (("text/plain", ("text", "plain", {}), 0.5)),
            ],
        )
        # compare mimetype precedence for a same given score
        self.assertEqual(
            parse_accept_header("audio/*, audio/basic"),
            [
                (("audio/basic", ("audio", "basic", {}), 1.0)),
                (("audio/*", ("audio", "*", {}), 1.0)),
            ],
        )
        self.assertEqual(
            parse_accept_header("text/*, text/html, text/html;level=1, */*"),
            [
                (("text/html", ("text", "html", {"level": "1"}), 1.0)),
                (("text/html", ("text", "html", {}), 1.0)),
                (("text/*", ("text", "*", {}), 1.0)),
                (("*/*", ("*", "*", {}), 1.0)),
            ],
        )
        # free party
        self.assertEqual(
            parse_accept_header(
                "text/*;q=0.3, text/html;q=0.7, text/html;level=1, text/html;level=2;q=0.4, */*;q=0.5"
            ),
            [
                (("text/html", ("text", "html", {"level": "1"}), 1.0)),
                (("text/html", ("text", "html", {}), 0.7)),
                (("*/*", ("*", "*", {}), 0.5)),
                (("text/html", ("text", "html", {"level": "2"}), 0.4)),
                (("text/*", ("text", "*", {}), 0.3)),
            ],
        )
        # chrome sample header
        self.assertEqual(
            parse_accept_header(
                "application/xml,application/xhtml+xml,text/html;q=0.9,text/plain;q=0.8,image/png,*/*;q=0.5"
            ),
            [
                (("application/xhtml+xml", ("application", "xhtml+xml", {}), 1.0)),
                (("application/xml", ("application", "xml", {}), 1.0)),
                (("image/png", ("image", "png", {}), 1.0)),
                (("text/html", ("text", "html", {}), 0.9)),
                (("text/plain", ("text", "plain", {}), 0.8)),
                (("*/*", ("*", "*", {}), 0.5)),
            ],
        )

    def test_parse_accept_language(self):
        self.assertEqual(
            _parse_accept_header("fr,fr-fr;q=0.8,en-us;q=0.5,en;q=0.3"),
            [
                ("fr", "fr", 1.0),
                ("fr-fr", "fr-fr", 0.8),
                ("en-us", "en-us", 0.5),
                ("en", "en", 0.3),
            ],
        )

    def test_parse_accept_charset(self):
        parse_accept_header = partial(
            _parse_accept_header, value_sort_key=_charset_sort_key
        )
        self.assertEqual(
            parse_accept_header("ISO-8859-1,utf-8;q=0.7,*;q=0.7"),
            [
                ("ISO-8859-1", "ISO-8859-1", 1.0),
                ("utf-8", "utf-8", 0.7),
                ("*", "*", 0.7),
            ],
        )


class WebRequestTC(unittest.TestCase):
    def test_negotiated_language(self):
        vreg = FakeCWRegistryStore(FakeConfig(), initlog=False)
        vreg.config.translations = {"fr": (None, None), "en": (None, None)}
        headers = {
            "Accept-Language": "fr,fr-fr;q=0.8,en-us;q=0.5,en;q=0.3",
        }
        req = CubicWebRequestBase(vreg, headers=headers)
        self.assertEqual(req.negotiated_language(), "fr")

    def test_build_url_language_from_url(self):
        vreg = FakeCWRegistryStore(FakeConfig(), initlog=False)
        vreg.config["base-url"] = BASE_URL
        vreg.config["language-mode"] = "url-prefix"
        vreg.config.translations["fr"] = str, str
        req = CubicWebRequestBase(vreg)
        # Override from_controller to avoid getting into relative_path method,
        # which is not implemented in CubicWebRequestBase.
        req.from_controller = lambda: "not view"
        self.assertEqual(req.lang, "en")  # site's default language
        self.assertEqual(req.build_url(), urljoin(BASE_URL, "en/view"))
        self.assertEqual(req.build_url("foo"), urljoin(BASE_URL, "en/foo"))
        req.set_language("fr")
        self.assertEqual(req.lang, "fr")
        self.assertEqual(req.build_url(), urljoin(BASE_URL, "fr/view"))
        self.assertEqual(req.build_url("foo"), urljoin(BASE_URL, "fr/foo"))
        # no language prefix in URL
        vreg.config["language-mode"] = ""
        self.assertEqual(req.build_url(), urljoin(BASE_URL, "view"))
        self.assertEqual(req.build_url("foo"), urljoin(BASE_URL, "foo"))
        req.set_language("fr")
        self.assertEqual(req.build_url(), urljoin(BASE_URL, "view"))
        self.assertEqual(req.build_url("foo"), urljoin(BASE_URL, "foo"))


class RequestCWTC(WebCWTC):
    def test_view_catch_ex(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('CWUser X WHERE X login "hop"')
            self.assertEqual(req.view("oneline", rset, "null"), "")
            self.assertRaises(ObjectNotFound, req.view, "onelinee", rset, "null")


if __name__ == "__main__":
    unittest.main()



from io import BytesIO

from logilab.common.testlib import TestCase, unittest_main

from cubicweb.devtools import htmlparser

from cubicweb_web.devtools.testlib import WebCWTC

HTML_PAGE = """<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html>
  <head><title>need a title</title></head>
  <body>
    <h1>Hello World !</h1>
  </body>
</html>
"""

HTML_PAGE2 = """<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html>
 <head><title>need a title</title></head>
 <body>
   <h1>Test</h1>
   <h1>Hello <a href="http://www.google.com">world</a> !</h1>
   <h2>h2 title</h2>
   <h3>h3 title</h3>
   <h2>antoher h2 title</h2>
   <h4>h4 title</h4>
   <p><a href="http://www.logilab.org">Logilab</a> introduces CW !</p>
 </body>
</html>
"""

HTML_PAGE_ERROR = """<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html>
 <head><title>need a title</title></head>
 <body>
   Logilab</a> introduces CW !
 </body>
</html>
"""

HTML_NON_STRICT = """<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html>
  <head><title>need a title</title></head>
  <body>
    <h1><a href="something.com">title</h1>
  </body>
</html>
"""


class HTMLPageInfoTC(TestCase):
    """test cases for PageInfo"""

    def setUp(self):
        parser = htmlparser.HTMLValidator()
        # disable cleanup that would remove doctype
        parser.preprocess_data = lambda data: data
        self.page_info = parser.parse_string(HTML_PAGE2)

    def test_source1(self):
        """make sure source is stored correctly"""
        self.assertEqual(self.page_info.source, HTML_PAGE2)

    def test_source2(self):
        """make sure source is stored correctly - raise exception"""
        parser = htmlparser.DTDValidator()
        self.assertRaises(AssertionError, parser.parse_string, HTML_PAGE_ERROR)

    def test_has_title_no_level(self):
        """tests h? tags information"""
        self.assertEqual(self.page_info.has_title("Test"), True)
        self.assertEqual(self.page_info.has_title("Test "), False)
        self.assertEqual(self.page_info.has_title("Tes"), False)
        self.assertEqual(self.page_info.has_title("Hello world !"), True)

    def test_has_title_level(self):
        """tests h? tags information"""
        self.assertEqual(self.page_info.has_title("Test", level=1), True)
        self.assertEqual(self.page_info.has_title("Test", level=2), False)
        self.assertEqual(self.page_info.has_title("Test", level=3), False)
        self.assertEqual(self.page_info.has_title("Test", level=4), False)
        self.assertRaises(IndexError, self.page_info.has_title, "Test", level=5)

    def test_has_title_regexp_no_level(self):
        """tests has_title_regexp() with no particular level specified"""
        self.assertEqual(self.page_info.has_title_regexp("h[23] title"), True)

    def test_has_title_regexp_level(self):
        """tests has_title_regexp() with a particular level specified"""
        self.assertEqual(self.page_info.has_title_regexp("h[23] title", 2), True)
        self.assertEqual(self.page_info.has_title_regexp("h[23] title", 3), True)
        self.assertEqual(self.page_info.has_title_regexp("h[23] title", 4), False)

    def test_appears(self):
        """tests PageInfo.appears()"""
        self.assertEqual(self.page_info.appears("CW"), True)
        self.assertEqual(self.page_info.appears("Logilab"), True)
        self.assertEqual(self.page_info.appears("Logilab introduces"), True)
        self.assertEqual(self.page_info.appears("H2 title"), False)

    def test_has_link(self):
        """tests has_link()"""
        self.assertEqual(self.page_info.has_link("Logilab"), True)
        self.assertEqual(self.page_info.has_link("logilab"), False)
        self.assertEqual(
            self.page_info.has_link("Logilab", "http://www.logilab.org"), True
        )
        self.assertEqual(
            self.page_info.has_link("Logilab", "http://www.google.com"), False
        )

    def test_has_link_regexp(self):
        """test has_link_regexp()"""
        self.assertEqual(self.page_info.has_link_regexp("L[oi]gilab"), True)
        self.assertEqual(self.page_info.has_link_regexp("L[ai]gilab"), False)


class FakeFormTC(TestCase):
    def test_fake_form(self):
        class entity:
            cw_etype = "Entity"
            eid = 0

        sio = BytesIO(b"hop\n")
        form = WebCWTC.fake_form(
            "import",
            {
                "file": ("filename.txt", sio),
                "encoding": "utf-8",
            },
            [(entity, {"field": "value"})],
        )
        self.assertEqual(
            form,
            {
                "__form_id": "import",
                "__maineid": 0,
                "__type:0": "Entity",
                "_cw_entity_fields:0": "__type,field",
                "_cw_fields": "encoding,file",
                "eid": [0],
                "encoding": "utf-8",
                "field:0": "value",
                "file": ("filename.txt", sio),
            },
        )


class WebRepoAccessTC(WebCWTC):
    def test_web_request(self):
        acc = self.new_access("admin")
        with acc.web_request(elephant="babar") as req:
            rset = req.execute("Any X WHERE X is CWUser")
            self.assertTrue(rset)
            self.assertEqual("babar", req.form["elephant"])


if __name__ == "__main__":
    unittest_main()



from os.path import join
from shutil import rmtree

from urllib.parse import urlparse, urljoin

from cubicweb.devtools import BASE_URL
from cubicweb_web.devtools.testlib import PyramidWebCWTC


class LoginTestLangUrlPrefix(PyramidWebCWTC):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.config.global_set_option("language-mode", "url-prefix")

    def test_login_password_login_lang_prefix(self):
        res = self.webapp.post(
            urljoin(BASE_URL, "fr/login"),
            {"__login": self.admlogin, "__password": self.admpassword},
        )
        self.assertEqual(res.status_int, 303)

        res = self.webapp.get(urljoin(BASE_URL, "fr/login"))
        self.assertEqual(res.status_int, 303)


class LoginTest(PyramidWebCWTC):
    def test_login_form(self):
        res = self.webapp.get(urljoin(BASE_URL, "login"))
        self.assertIn("__login", res.text)

    def test_login_password_login(self):
        res = self.webapp.post(
            urljoin(BASE_URL, "login"),
            {"__login": self.admlogin, "__password": self.admpassword},
        )
        self.assertEqual(res.status_int, 303)

        res = self.webapp.get(urljoin(BASE_URL, "login"))
        self.assertEqual(res.status_int, 303)

    def test_login_password_login_cookie_expires(self):
        res = self.webapp.post(
            urljoin(BASE_URL, "login"),
            {"__login": self.admlogin, "__password": self.admpassword},
        )
        self.assertEqual(res.status_int, 303)

        cookies = self.webapp.cookiejar._cookies[urlparse(BASE_URL).netloc]["/"]
        self.assertNotIn("pauth_tkt", cookies)
        self.assertIn("auth_tkt", cookies)
        self.assertIsNone(cookies["auth_tkt"].expires)

        res = self.webapp.logout()
        self.assertEqual(res.status_int, 303)

        self.assertNotIn("auth_tkt", cookies)
        self.assertNotIn("pauth_tkt", cookies)

        res = self.webapp.post(
            urljoin(BASE_URL, "login"),
            {
                "__login": self.admlogin,
                "__password": self.admpassword,
                "__setauthcookie": 1,
            },
        )
        self.assertEqual(res.status_int, 303)

        cookies = self.webapp.cookiejar._cookies[urlparse(BASE_URL).netloc]["/"]
        self.assertNotIn("auth_tkt", cookies)
        self.assertIn("pauth_tkt", cookies)
        self.assertIsNotNone(cookies["pauth_tkt"].expires)

    def test_login_bad_password(self):
        self.config.i18ncompile(["en", "fr"])
        try:
            self.config._gettext_init()
            res = self.webapp.post(
                urljoin(BASE_URL, "login"),
                {"__login": self.admlogin, "__password": "empty"},
                headers={"Accept-Language": "fr"},
                status=403,
            )
        finally:
            rmtree(join(self.config.apphome, "i18n"))
        self.assertIn("\xc9chec de l&#39;authentification", res.text)

    def test_same_site_lax_by_default(self):
        res = self.webapp.post(
            urljoin(BASE_URL, "login"),
            {"__login": self.admlogin, "__password": self.admpassword},
        )
        for cookie in res.headers.getall("Set-Cookie"):
            self.assertIn("SameSite=Lax", cookie)
        self.assertTrue(len(res.headers.getall("Set-Cookie")) > 0)


class LoginRedirectionTC(PyramidWebCWTC):
    anonymous_allowed = False

    def setUp(self):
        self.route_prefix = urlparse(BASE_URL).path
        return super().setUp()

    def test_redirect_to_cubicweb_page_with_arguments(self):
        res = self.webapp.get(urljoin(BASE_URL, "cwsource?args=42"), status=303)
        self.assertIn(
            f"/login?postlogin_path={self.route_prefix}cwsource%3Fargs%3D42", res.text
        )

    def test_redirect_to_cubicweb_page_without_arguments(self):
        res = self.webapp.get(urljoin(BASE_URL, "cwsource"), status=303)
        self.assertIn(f"/login?postlogin_path={self.route_prefix}cwsource", res.text)

    def test_redirect_to_main_page_with_arguments(self):
        res = self.webapp.get(urljoin(BASE_URL, "?args=42"), status=303)
        self.assertIn(
            f"/login?postlogin_path={self.route_prefix}%3Fargs%3D42", res.text
        )


class CookieParametersTC(PyramidWebCWTC):
    settings = {
        **PyramidWebCWTC.settings,
        "cubicweb.auth.authtkt.session.samesite": "None",
    }

    def test_same_site_set_from_config(self):
        res = self.webapp.post(
            urljoin(BASE_URL, "login"),
            {"__login": self.admlogin, "__password": self.admpassword},
        )
        for cookie in res.headers.getall("Set-Cookie"):
            self.assertIn("SameSite=None", cookie)
        self.assertTrue(len(res.headers.getall("Set-Cookie")) > 0)


if __name__ == "__main__":
    from unittest import main

    main()



from unittest import skipUnless

from cubicweb_web.devtools.testlib import WebCWTC

try:
    import docutils  # noqa

    DOCUTILS_AVAILABLE = True
except ImportError:
    DOCUTILS_AVAILABLE = False


class WdocViewsCWTC(WebCWTC):
    @skipUnless(
        DOCUTILS_AVAILABLE, "rest_publish does not support ..winclude without docutils"
    )
    def test(self):
        with self.admin_access.web_request(fid="main") as req:
            page = req.view("wdoc")
        self.assertIn("Site documentation", page)
        # This part is renderend through rst extension (..winclude directive).
        self.assertIn(
            "This web application is based on the CubicWeb knowledge management system",
            page,
        )


if __name__ == "__main__":
    import unittest

    unittest.main()



from cubicweb._exceptions import UnknownEid


class HasCWPermissionRoutePredicate:
    """A route predicate that matches if the current user has given permission
    on the entity given by its eid, using cubicweb.predicates.has_permission.

    The main purpose of this predicate is to restrain access on specific pages,
    based on the permission of the current user.
    """

    def __init__(self, permission, config):
        self.permission = permission

    def text(self):
        return f"has_cw_permission = {self.permission}"

    phash = text

    def __call__(self, info, request):
        eid = self._get_eid(info, request)
        if not eid:
            return False

        if request.cw_cnx is None:
            return False

        try:
            entity = request.cw_cnx.entity_from_eid(eid)
        except UnknownEid:
            return False

        return entity.cw_has_perm(self.permission)

    def _get_eid(self, info, request):
        try:
            eid = int(info["match"]["eid"])
        except (KeyError, ValueError):
            return None
        return eid


class HasCWPermissionViewPredicate(HasCWPermissionRoutePredicate):
    """This predicate is quite the same as its route pendant, but for
    view.

    The main purpose of this predicate is to have different views for a
    single route, but with different features according to current user
    permissions on the entity.
    """

    def _get_eid(self, info, request):
        try:
            eid = int(request.matchdict["eid"])
        except (KeyError, ValueError):
            return None
        return eid


def includeme(config):
    config.add_route_predicate("has_cw_permission", HasCWPermissionRoutePredicate)
    config.add_view_predicate("has_cw_permission", HasCWPermissionViewPredicate)



# copyright 2012-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.


from logilab.common.registry import Predicate

from cubicweb import UnknownEid, tags, transaction as tx
from cubicweb import _
from cubicweb.predicates import match_kwargs
from cubicweb.schema import display_name
from cubicweb_web.view import View, StartupView


class undoable_action(Predicate):
    """Select only undoable actions depending on filters provided. Undo Action
    is expected to be specified by the `tx_action` argument.

    Currently the only implemented filter is:

    :param action_type: chars among CUDAR (standing for Create, Update, Delete,
                        Add, Remove)
    """

    # XXX FIXME : this selector should be completed to allow selection on the
    # entity or relation types and public / private.
    def __init__(self, action_type="CUDAR"):
        assert not set(action_type) - set("CUDAR")
        self.action_type = action_type

    def __str__(self):
        return "%s()" % (self.__class__.__name__)

    def __call__(self, cls, req, tx_action=None, **kwargs):
        # tx_action is expected to be a transaction.AbstractAction
        if not isinstance(tx_action, tx.AbstractAction):
            return 0
        # Filter according to action type
        return int(tx_action.action in self.action_type)


class UndoHistoryView(StartupView):
    __regid__ = "undohistory"
    title = _("Undoing")
    item_vid = "undoable-transaction-view"
    cache_max_age = 0

    redirect_path = "view"  # TODO
    redirect_params = dict(vid="undohistory")  # TODO
    public_actions_only = True

    # TODO Allow to choose if if want all actions or only the public ones
    # (default)

    def call(self, **kwargs):
        txs = self._cw.cnx.undoable_transactions()
        if txs:
            self.w("<ul class='undo-transactions'>")
            for a_tx in txs:
                self.cell_call(a_tx)
            self.w("</ul>")

    def cell_call(self, tx):
        self.w("<li>")
        self.wview(
            self.item_vid,
            None,
            txuuid=tx.uuid,
            public=self.public_actions_only,
            redirect_path=self.redirect_path,
            redirect_params=self.redirect_params,
        )
        self.w("</li>\n")


class UndoableTransactionView(View):
    __regid__ = "undoable-transaction-view"
    __select__ = View.__select__ & match_kwargs("txuuid")

    item_vid = "undoable-action-list-view"
    cache_max_age = 0

    def build_undo_link(self, txuuid, redirect_path=None, redirect_params=None):
        """the kwargs are passed to build_url"""
        _ = self._cw._
        redirect = {}
        if redirect_path:
            redirect["__redirectpath"] = redirect_path
        if redirect_params:
            if isinstance(redirect_params, dict):
                redirect["__redirectparams"] = self._cw.build_url_params(
                    **redirect_params
                )
            else:
                redirect["__redirectparams"] = redirect_params
        link_url = self._cw.build_url("undo", txuuid=txuuid, **redirect)
        msg = "<span class='undo'>%s</span>" % tags.a(_("undo"), href=link_url)
        return msg

    def call(self, txuuid, public=True, redirect_path=None, redirect_params=None):
        _ = self._cw._
        txinfo = self._cw.cnx.transaction_info(txuuid)
        try:
            # XXX Under some unknown circumstances txinfo.user_eid=-1
            user = self._cw.entity_from_eid(txinfo.user_eid)
        except UnknownEid:
            user = None
        undo_url = self.build_undo_link(
            txuuid, redirect_path=redirect_path, redirect_params=redirect_params
        )
        txinfo_dict = dict(
            dt=self._cw.format_date(txinfo.datetime, time=True),
            user_eid=txinfo.user_eid,
            user=user and user.view("outofcontext") or _("undefined user"),
            txuuid=txuuid,
            undo_link=undo_url,
        )
        self.w(_("By %(user)s on %(dt)s [%(undo_link)s]"), txinfo_dict, escape=False)

        tx_actions = txinfo.actions_list(public=public)
        if tx_actions:
            self.wview(self.item_vid, None, tx_actions=tx_actions)


class UndoableActionListView(View):
    __regid__ = "undoable-action-list-view"
    __select__ = View.__select__ & match_kwargs("tx_actions")
    title = _("Undoable actions")
    item_vid = "undoable-action-view"
    cache_max_age = 0

    def call(self, tx_actions):
        if tx_actions:
            self.w("<ol class='undo-actions'>")
            for action in tx_actions:
                self.cell_call(action)
            self.w("</ol>")

    def cell_call(self, action):
        self.w("<li>")
        self.wview(self.item_vid, None, tx_action=action)
        self.w("</li>\n")


class UndoableActionBaseView(View):
    __regid__ = "undoable-action-view"
    __abstract__ = True

    def call(self, tx_action):
        raise NotImplementedError(self)

    def _build_entity_link(self, eid):
        try:
            entity = self._cw.entity_from_eid(eid)
            return entity.view("outofcontext")
        except UnknownEid:
            return _("(suppressed) entity #%d") % eid

    def _build_relation_info(self, rtype, eid_from, eid_to):
        return dict(
            rtype=display_name(self._cw, rtype),
            entity_from=self._build_entity_link(eid_from),
            entity_to=self._build_entity_link(eid_to),
        )

    def _build_entity_info(self, etype, eid, changes):
        return dict(
            etype=display_name(self._cw, etype),
            entity=self._build_entity_link(eid),
            eid=eid,
            changes=changes,
        )


class UndoableAddActionView(UndoableActionBaseView):
    __select__ = UndoableActionBaseView.__select__ & undoable_action(action_type="A")

    def call(self, tx_action):
        _ = self._cw._
        self.w(
            _("Added relation : %(entity_from)s %(rtype)s %(entity_to)s"),
            self._build_relation_info(
                tx_action.rtype, tx_action.eid_from, tx_action.eid_to
            ),
            escape=False,
        )


class UndoableRemoveActionView(UndoableActionBaseView):
    __select__ = UndoableActionBaseView.__select__ & undoable_action(action_type="R")

    def call(self, tx_action):
        _ = self._cw._
        self.w(
            _("Delete relation : %(entity_from)s %(rtype)s %(entity_to)s"),
            self._build_relation_info(
                tx_action.rtype, tx_action.eid_from, tx_action.eid_to
            ),
            escape=False,
        )


class UndoableCreateActionView(UndoableActionBaseView):
    __select__ = UndoableActionBaseView.__select__ & undoable_action(action_type="C")

    def call(self, tx_action):
        _ = self._cw._
        self.w(
            _("Created %(etype)s : %(entity)s"),
            self._build_entity_info(tx_action.etype, tx_action.eid, tx_action.changes),
            escape=False,
        )


class UndoableDeleteActionView(UndoableActionBaseView):
    __select__ = UndoableActionBaseView.__select__ & undoable_action(action_type="D")

    def call(self, tx_action):
        _ = self._cw._
        self.w(
            _("Deleted %(etype)s : %(entity)s"),
            self._build_entity_info(tx_action.etype, tx_action.eid, tx_action.changes),
            escape=False,
        )


class UndoableUpdateActionView(UndoableActionBaseView):
    __select__ = UndoableActionBaseView.__select__ & undoable_action(action_type="U")

    def call(self, tx_action):
        _ = self._cw._
        self.w(
            _("Updated %(etype)s : %(entity)s"),
            self._build_entity_info(tx_action.etype, tx_action.eid, tx_action.changes),
            escape=False,
        )



# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""
mainly regression-preventing tests for reledit views
"""
from cubicweb.devtools import BASE_URL

from cubicweb_web.devtools.testlib import WebCWTC
from cubicweb_web.views.uicfg import reledit_ctrl


class ReleditMixinTC:
    def setup_database(self):
        with self.admin_access.client_cnx() as cnx:
            self.proj = cnx.create_entity(
                "Project", title="cubicweb-world-domination"
            ).eid
            self.tick = cnx.create_entity("Ticket", title="write the code").eid
            self.toto = cnx.create_entity("Personne", nom="Toto").eid
            cnx.commit()


class ClickAndEditFormCWTC(ReleditMixinTC, WebCWTC):
    def setup_database(self):
        super().setup_database()
        with self.admin_access.client_cnx() as cnx:
            cnx.create_entity("CWGroup", name="validators")
            cnx.create_entity("CWGroup", name="contributors")

            self.create_user(
                cnx, login="edith", password="edith", groups=("validators", "users")
            )

            self.create_user(
                cnx,
                login="contrib",
                password="contrib",
                groups=("contributors", "users"),
            )
            cnx.commit()

    @property
    def editable_fields(self):
        return {
            "title": f"""<div id="title-subject-%(eid)s-reledit" onmouseout="jQuery('#title-subject-%(eid)s').addClass('invisible')" onmouseover="jQuery('#title-subject-%(eid)s').removeClass('invisible')" class="releditField"><div id="title-subject-%(eid)s-value" class="editableFieldValue">cubicweb-world-domination</div><div id="title-subject-%(eid)s" class="editableField invisible"><div id="title-subject-%(eid)s-update" class="editableField" onclick="cw.reledit.loadInlineEditionForm(&#39;base&#39;, %(eid)s, &#39;title&#39;, &#39;subject&#39;, &#39;title-subject-%(eid)s&#39;, false, &#39;&#39;, &#39;edit_rtype&#39;);" title="click to edit this field"><img title="click to edit this field" src="{BASE_URL}data/pen_icon.png" alt="click to edit this field"/></div></div></div>""",
            "long_desc": f"""<div id="long_desc-subject-%(eid)s-reledit" onmouseout="jQuery('#long_desc-subject-%(eid)s').addClass('invisible')" onmouseover="jQuery('#long_desc-subject-%(eid)s').removeClass('invisible')" class="releditField"><div id="long_desc-subject-%(eid)s-value" class="editableFieldValue">&lt;not specified&gt;</div><div id="long_desc-subject-%(eid)s" class="editableField invisible"><div id="long_desc-subject-%(eid)s-add" class="editableField" onclick="cw.reledit.loadInlineEditionForm(&#39;edition&#39;, %(eid)s, &#39;long_desc&#39;, &#39;subject&#39;, &#39;long_desc-subject-%(eid)s&#39;, false, &#39;autolimited&#39;, &#39;add&#39;);" title="click to add a value"><img title="click to add a value" src="{BASE_URL}data/plus.png" alt="click to add a value"/></div></div></div>""",
            "manager": f"""<div id="manager-subject-%(eid)s-reledit" onmouseout="jQuery('#manager-subject-%(eid)s').addClass('invisible')" onmouseover="jQuery('#manager-subject-%(eid)s').removeClass('invisible')" class="releditField"><div id="manager-subject-%(eid)s-value" class="editableFieldValue">&lt;not specified&gt;</div><div id="manager-subject-%(eid)s" class="editableField invisible"><div id="manager-subject-%(eid)s-update" class="editableField" onclick="cw.reledit.loadInlineEditionForm(&#39;base&#39;, %(eid)s, &#39;manager&#39;, &#39;subject&#39;, &#39;manager-subject-%(eid)s&#39;, false, &#39;autolimited&#39;, &#39;edit_rtype&#39;);" title="click to edit this field"><img title="click to edit this field" src="{BASE_URL}data/pen_icon.png" alt="click to edit this field"/></div></div></div>""",
            "composite_card11_2ttypes": """&lt;not specified&gt;""",
            "concerns": """&lt;not specified&gt;""",
        }

    def _test_fields(self, fields, login="admin"):
        with self.new_access(login).web_request() as req:
            proj = req.entity_from_eid(self.proj)
            for rschema, ttypes, role in proj.e_schema.relation_definitions(
                include_final=True
            ):
                if rschema not in fields:
                    continue
                rtype = rschema.type
                proj_view = proj.view("reledit", rtype=rtype, role=role)
                if proj_view:
                    self.assertMultiLineEqual(
                        fields[rtype] % {"eid": self.proj}, proj_view, rtype
                    )

    def test_admins_edit_permissions(self):
        """managers can edit all fields"""
        fields = self.editable_fields.copy()
        self._test_fields(fields, login="admin")

    def test_validators_edit_permissions(self):
        """validators can edit all relations"""
        fields = self.editable_fields.copy()
        self._test_fields(fields, login="edith")

    def test_contributors_edit_permissions(self):
        """contributors can edit all but `manager` relation as they don't update
        permissions on Personne
        """
        fields = self.editable_fields.copy()
        fields["manager"] = """&lt;not specified&gt;"""
        self._test_fields(fields, login="contrib")

    def test_admins_edit_permissions_multiple(self):
        """
        Add test for the permission of relation for all subjects possibles.
        - Personne has the relation "tags" only with "BlogEntry" but
        - Tag could have the relations "tags" with "BlogEntry" and "CWUser"
        This test check that do not generate the KeyError when trying to get
        the relation between "Personne" <--> "CWUser"
        """
        with self.new_access("admin").web_request() as cnx:
            entity = cnx.find("Personne").limit(1).one()
            entity.view("reledit", rtype="tags", role="subject")

    def test_default_forms(self):
        self.skipTest(
            "Need to check if this test should still run post reledit/doreledit merge"
        )
        doreledit = {
            "title": f"""<div id="title-subject-%(eid)s-reledit" onmouseout="jQuery('#title-subject-%(eid)s').addClass('invisible')" onmouseover="jQuery('#title-subject-%(eid)s').removeClass('invisible')" class="releditField"><div id="title-subject-%(eid)s-value" class="editableFieldValue">cubicweb-world-domination</div><form action="{BASE_URL}validateform?__onsuccess=window.parent.cw.reledit.onSuccess" method="post" enctype="application/x-www-form-urlencoded" id="title-subject-%(eid)s-form" onsubmit="return freezeFormButtons(&#39;title-subject-%(eid)s-form&#39;);" class="releditForm" target="eformframe">
<fieldset>
<input name="__form_id" type="hidden" value="base" />
<input name="__errorurl" type="hidden" value="{BASE_URL}view?rql=Blop&amp;vid=blop#title-subject-%(eid)s-form" />
<input name="__domid" type="hidden" value="title-subject-%(eid)s-form" />
<input name="__type:%(eid)s" type="hidden" value="Project" />
<input name="eid" type="hidden" value="%(eid)s" />
<input name="__maineid" type="hidden" value="%(eid)s" />
<input name="__reledit|vid" type="hidden" value="" />
<input name="__reledit|rtype" type="hidden" value="title" />
<input name="__reledit|divid" type="hidden" value="title-subject-%(eid)s" />
<input name="__reledit|formid" type="hidden" value="base" />
<input name="__reledit|reload" type="hidden" value="false" />
<input name="__reledit|role" type="hidden" value="subject" />
<input name="__reledit|eid" type="hidden" value="%(eid)s" />
<input name="_cw_entity_fields:%(eid)s" type="hidden" value="title-subject,__type" />
<fieldset class="default">
<table class="">
<tr class="title_subject_row">
<td>
<input id="title-subject:%(eid)s" maxlength="32" name="title-subject:%(eid)s" size="32" type="text" value="cubicweb-world-domination" />
</td></tr>
</table></fieldset>
<table class="buttonbar">
<tr>
<td><button class="validateButton" type="submit" value="button_ok"><img alt="OK_ICON" src="{BASE_URL}data/ok.png" />button_ok</button></td>
<td><button class="validateButton" onclick="cw.reledit.cleanupAfterCancel(&#39;title-subject-%(eid)s&#39;)" type="button" value="button_cancel"><img alt="CANCEL_ICON" src="{BASE_URL}data/cancel.png" />button_cancel</button></td>
</tr></table>
</fieldset>
<iframe width="0px" height="0px" src="javascript: void(0);" name="eformframe" id="eformframe"></iframe>
</form><div id="title-subject-%(eid)s" class="editableField invisible"><div id="title-subject-%(eid)s-update" class="editableField" onclick="cw.reledit.loadInlineEditionForm(&#39;base&#39;, %(eid)s, &#39;title&#39;, &#39;subject&#39;, &#39;title-subject-%(eid)s&#39;, false, &#39;&#39;);" title="click to edit this field"><img title="click to edit this field" src="{BASE_URL}data/pen_icon.png" alt="click to edit this field"/></div></div></div>""",
            "long_desc": f"""<div id="long_desc-subject-%(eid)s-reledit" onmouseout="jQuery('#long_desc-subject-%(eid)s').addClass('invisible')" onmouseover="jQuery('#long_desc-subject-%(eid)s').removeClass('invisible')" class="releditField"><div id="long_desc-subject-%(eid)s-value" class="editableFieldValue">&lt;not specified&gt;</div><form action="{BASE_URL}validateform?__onsuccess=window.parent.cw.reledit.onSuccess" method="post" enctype="application/x-www-form-urlencoded" id="long_desc-subject-%(eid)s-form" onsubmit="return freezeFormButtons(&#39;long_desc-subject-%(eid)s-form&#39;);" class="releditForm" target="eformframe">
<fieldset>
<input name="__form_id" type="invisible" value="edition" />
<input name="__errorurl" type="invisible" value="{BASE_URL}view?rql=Blop&amp;vid=blop#long_desc-subject-%(eid)s-form" />
<input name="__domid" type="hidden" value="long_desc-subject-%(eid)s-form" />
<input name="__type:A" type="hidden" value="Blog" />
<input name="eid" type="hidden" value="A" />
<input name="__maineid" type="hidden" value="A" />
<input name="__linkto" type="hidden" value="long_desc:%(eid)s:object" />
<input name="__message" type="hidden" value="entity linked" />
<input name="__reledit|vid" type="hidden" value="autolimited" />
<input name="__reledit|rtype" type="hidden" value="long_desc" />
<input name="__reledit|divid" type="hidden" value="long_desc-subject-%(eid)s" />
<input name="__reledit|formid" type="hidden" value="edition" />
<input name="__reledit|reload" type="hidden" value="false" />
<input name="__reledit|role" type="hidden" value="subject" />
<input name="__reledit|eid" type="hidden" value="%(eid)s" />
<input name="_cw_entity_fields:A" type="hidden" value="title-subject,rss_url-subject,__type,description-subject" />
<fieldset class="default">
<table class="attributeForm">
<tr class="title_subject_row">
<th class="labelCol"><label class="required" for="title-subject:A">title</label></th>
<td>
<input id="title-subject:A" maxlength="50" name="title-subject:A" size="45" type="text" value="" />
</td></tr>
<tr class="description_subject_row">
<th class="labelCol"><label for="description-subject:A">description</label></th>
<td>
<input name="description_format-subject:A" type="hidden" value="text/html" /><textarea cols="80" cubicweb:type="wysiwyg" id="description-subject:A" name="description-subject:A" onkeyup="autogrow(this)" rows="2"></textarea>
</td></tr>
<tr class="rss_url_subject_row">
<th class="labelCol"><label for="rss_url-subject:A">rss_url</label></th>
<td>
<input id="rss_url-subject:A" maxlength="128" name="rss_url-subject:A" size="45" type="text" value="" />
</td></tr>
</table></fieldset>
<table class="buttonbar">
<tr>
<td><button class="validateButton" type="submit" value="button_ok"><img alt="OK_ICON" src="{BASE_URL}data/ok.png" />button_ok</button></td>
<td><button class="validateButton" onclick="cw.reledit.cleanupAfterCancel(&#39;long_desc-subject-%(eid)s&#39;)" type="button" value="button_cancel"><img alt="CANCEL_ICON" src="{BASE_URL}data/cancel.png" />button_cancel</button></td>
</tr></table>
</fieldset>
<iframe width="0px" height="0px" src="javascript: void(0);" name="eformframe" id="eformframe"></iframe>
</form><div id="long_desc-subject-%(eid)s" class="editableField invisible"><div id="long_desc-subject-%(eid)s-add" class="editableField" onclick="cw.reledit.loadInlineEditionForm(&#39;edition&#39;, %(eid)s, &#39;long_desc&#39;, &#39;subject&#39;, &#39;long_desc-subject-%(eid)s&#39;, false, &#39;autolimited&#39;);" title="click to add a value"><img title="click to add a value" src="{BASE_URL}data/plus.png" alt="click to add a value"/></div></div></div>""",
            "manager": f"""<div id="manager-subject-%(eid)s-reledit" onmouseout="jQuery('#manager-subject-%(eid)s').addClass('invisible')" onmouseover="jQuery('#manager-subject-%(eid)s').removeClass('invisible')" class="releditField"><div id="manager-subject-%(eid)s-value" class="editableFieldValue">&lt;not specified&gt;</div><form action="{BASE_URL}validateform?__onsuccess=window.parent.cw.reledit.onSuccess" method="post" enctype="application/x-www-form-urlencoded" id="manager-subject-%(eid)s-form" onsubmit="return freezeFormButtons(&#39;manager-subject-%(eid)s-form&#39;);" class="releditForm" target="eformframe">
<fieldset>
<input name="__form_id" type="hidden" value="base" />
<input name="__errorurl" type="hidden" value="{BASE_URL}view?rql=Blop&amp;vid=blop#manager-subject-%(eid)s-form" />
<input name="__domid" type="hidden" value="manager-subject-%(eid)s-form" />
<input name="__type:%(eid)s" type="hidden" value="Project" />
<input name="eid" type="hidden" value="%(eid)s" />
<input name="__maineid" type="hidden" value="%(eid)s" />
<input name="__linkto" type="hidden" value="long_desc:%(eid)s:object" />
<input name="__message" type="hidden" value="entity linked" />
<input name="__reledit|vid" type="hidden" value="autolimited" />
<input name="__reledit|rtype" type="hidden" value="manager" />
<input name="__reledit|divid" type="hidden" value="manager-subject-%(eid)s" />
<input name="__reledit|formid" type="hidden" value="base" />
<input name="__reledit|reload" type="hidden" value="false" />
<input name="__reledit|role" type="hidden" value="subject" />
<input name="__reledit|eid" type="hidden" value="%(eid)s" />
<input name="_cw_entity_fields:%(eid)s" type="hidden" value="manager-subject,__type" />
<fieldset class="default">
<table class="">
<tr class="manager_subject_row">
<td>
<select id="manager-subject:%(eid)s" name="manager-subject:%(eid)s" size="1">
<option value="__cubicweb_internal_field__"></option>
<option value="%(toto)s">Toto</option>
</select>
</td></tr>
</table></fieldset>
<table class="buttonbar">
<tr>
<td><button class="validateButton" type="submit" value="button_ok"><img alt="OK_ICON" src="{BASE_URL}data/ok.png" />button_ok</button></td>
<td><button class="validateButton" onclick="cw.reledit.cleanupAfterCancel(&#39;manager-subject-%(eid)s&#39;)" type="button" value="button_cancel"><img alt="CANCEL_ICON" src="{BASE_URL}data/cancel.png" />button_cancel</button></td>
</tr></table>
</fieldset>
<iframe width="0px" height="0px" src="javascript: void(0);" name="eformframe" id="eformframe"></iframe>
</form><div id="manager-subject-%(eid)s" class="editableField invisible"><div id="manager-subject-%(eid)s-update" class="editableField" onclick="cw.reledit.loadInlineEditionForm(&#39;base&#39;, %(eid)s, &#39;manager&#39;, &#39;subject&#39;, &#39;manager-subject-%(eid)s&#39;, false, &#39;autolimited&#39;);" title="click to edit this field"><img title="click to edit this field" src="{BASE_URL}data/pen_icon.png" alt="click to edit this field"/></div></div></div>""",
            "composite_card11_2ttypes": """&lt;not specified&gt;""",
            "concerns": """&lt;not specified&gt;""",
        }
        for rschema, ttypes, role in self.proj.e_schema.relation_definitions(
            includefinal=True
        ):
            if rschema not in doreledit:
                continue
            rtype = rschema.type
            self.assertMultiLineEqual(
                doreledit[rtype] % {"eid": self.proj.eid, "toto": self.toto.eid},
                self.proj.view(
                    "doreledit",
                    rtype=rtype,
                    role=role,
                    formid="edition" if rtype == "long_desc" else "base",
                ),
                rtype,
            )


class ClickAndEditFormUICFGTC(ReleditMixinTC, WebCWTC):
    def setup_database(self):
        super().setup_database()
        with self.admin_access.client_cnx() as cnx:
            cnx.execute(
                "SET T concerns P WHERE T eid %(t)s, P eid %(p)s",
                {"t": self.tick, "p": self.proj},
            )
            cnx.execute(
                "SET P manager T WHERE P eid %(p)s, T eid %(t)s",
                {"p": self.proj, "t": self.toto},
            )
            cnx.commit()

    def test_with_uicfg(self):
        old_rctl = reledit_ctrl._tagdefs.copy()
        reledit_ctrl.tag_attribute(
            ("Project", "title"),
            {"novalue_label": "<title is required>", "reload": True},
        )
        reledit_ctrl.tag_subject_of(
            ("Project", "long_desc", "*"),
            {
                "reload": True,
                "edit_target": "rtype",
                "novalue_label": "<long_desc is required>",
            },
        )
        reledit_ctrl.tag_subject_of(
            ("Project", "manager", "*"), {"edit_target": "related"}
        )
        reledit_ctrl.tag_subject_of(
            ("Project", "composite_card11_2ttypes", "*"), {"edit_target": "related"}
        )
        reledit_ctrl.tag_object_of(
            ("Ticket", "concerns", "Project"), {"edit_target": "rtype"}
        )
        reledit = {
            "title": f"""<div id="title-subject-%(eid)s-reledit" onmouseout="jQuery('#title-subject-%(eid)s').addClass('invisible')" onmouseover="jQuery('#title-subject-%(eid)s').removeClass('invisible')" class="releditField"><div id="title-subject-%(eid)s-value" class="editableFieldValue">cubicweb-world-domination</div><div id="title-subject-%(eid)s" class="editableField invisible"><div id="title-subject-%(eid)s-update" class="editableField" onclick="cw.reledit.loadInlineEditionForm(&#39;base&#39;, %(eid)s, &#39;title&#39;, &#39;subject&#39;, &#39;title-subject-%(eid)s&#39;, true, &#39;&#39;, &#39;edit_rtype&#39;);" title="click to edit this field"><img title="click to edit this field" src="{BASE_URL}data/pen_icon.png" alt="click to edit this field"/></div></div></div>""",
            "long_desc": f"""<div id="long_desc-subject-%(eid)s-reledit" onmouseout="jQuery('#long_desc-subject-%(eid)s').addClass('invisible')" onmouseover="jQuery('#long_desc-subject-%(eid)s').removeClass('invisible')" class="releditField"><div id="long_desc-subject-%(eid)s-value" class="editableFieldValue">&lt;long_desc is required&gt;</div><div id="long_desc-subject-%(eid)s" class="editableField invisible"><div id="long_desc-subject-%(eid)s-update" class="editableField" onclick="cw.reledit.loadInlineEditionForm(&#39;base&#39;, %(eid)s, &#39;long_desc&#39;, &#39;subject&#39;, &#39;long_desc-subject-%(eid)s&#39;, true, &#39;autolimited&#39;, &#39;edit_rtype&#39;);" title="click to edit this field"><img title="click to edit this field" src="{BASE_URL}data/pen_icon.png" alt="click to edit this field"/></div></div></div>""",
            "manager": f"""<div id="manager-subject-%(eid)s-reledit" onmouseout="jQuery('#manager-subject-%(eid)s').addClass('invisible')" onmouseover="jQuery('#manager-subject-%(eid)s').removeClass('invisible')" class="releditField"><div id="manager-subject-%(eid)s-value" class="editableFieldValue"><a href="{BASE_URL}personne/%(toto)s" title="">Toto</a></div><div id="manager-subject-%(eid)s" class="editableField invisible"><div id="manager-subject-%(eid)s-update" class="editableField" onclick="cw.reledit.loadInlineEditionForm(&#39;edition&#39;, %(eid)s, &#39;manager&#39;, &#39;subject&#39;, &#39;manager-subject-%(eid)s&#39;, false, &#39;autolimited&#39;, &#39;edit_related&#39;);" title="click to edit this field"><img title="click to edit this field" src="{BASE_URL}data/pen_icon.png" alt="click to edit this field"/></div><div id="manager-subject-%(eid)s-delete" class="editableField" onclick="cw.reledit.loadInlineEditionForm(&#39;deleteconf&#39;, %(eid)s, &#39;manager&#39;, &#39;subject&#39;, &#39;manager-subject-%(eid)s&#39;, false, &#39;autolimited&#39;, &#39;delete&#39;);" title="click to delete this value"><img title="click to delete this value" src="{BASE_URL}data/cancel.png" alt="click to delete this value"/></div></div></div>""",
            "composite_card11_2ttypes": """&lt;not specified&gt;""",
            "concerns": f"""<div id="concerns-object-%(eid)s-reledit" onmouseout="jQuery('#concerns-object-%(eid)s').addClass('invisible')" onmouseover="jQuery('#concerns-object-%(eid)s').removeClass('invisible')" class="releditField"><div id="concerns-object-%(eid)s-value" class="editableFieldValue"><a href="{BASE_URL}ticket/%(tick)s" title="">write the code</a></div><div id="concerns-object-%(eid)s" class="editableField invisible"><div id="concerns-object-%(eid)s-update" class="editableField" onclick="cw.reledit.loadInlineEditionForm(&#39;base&#39;, %(eid)s, &#39;concerns&#39;, &#39;object&#39;, &#39;concerns-object-%(eid)s&#39;, false, &#39;autolimited&#39;, &#39;edit_rtype&#39;);" title="click to edit this field"><img title="click to edit this field" src="{BASE_URL}data/pen_icon.png" alt="click to edit this field"/></div></div></div>""",
        }
        with self.admin_access.web_request() as req:
            proj = req.entity_from_eid(self.proj)
            for rschema, ttypes, role in proj.e_schema.relation_definitions(
                include_final=True
            ):
                if rschema not in reledit:
                    continue
                rtype = rschema.type
                self.assertMultiLineEqual(
                    reledit[rtype]
                    % {
                        "eid": self.proj,
                        "toto": req.cnx.entity_from_eid(self.toto).nom,
                        "tick": self.tick,
                    },
                    proj.view("reledit", rtype=rtype, role=role),
                    rtype,
                )
        reledit_ctrl.clear()
        reledit_ctrl._tagdefs.update(old_rctl)


if __name__ == "__main__":
    from logilab.common.testlib import unittest_main

    unittest_main()



# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

import time
from datetime import datetime

import pytz
from lxml import html

from cubicweb import Binary, ValidationError
from cubicweb.devtools import BASE_URL

from cubicweb_web.devtools.testlib import WebCWTC
from cubicweb_web.formfields import (
    IntField,
    StringField,
    RichTextField,
    PasswordField,
    DateTimeField,
    FileField,
    EditableFileField,
    TZDatetimeField,
)
from cubicweb_web.formwidgets import DateTimePicker, JQueryDateTimePicker
from cubicweb_web.views.formrenderers import FormRenderer
from cubicweb_web.views.forms import EntityFieldsForm, FieldsForm
from cubicweb_web.views.workflow import ChangeStateForm


class FieldsFormCWTC(WebCWTC):
    def test_form_field_format(self):
        with self.admin_access.web_request() as req:
            form = FieldsForm(req, None)
            self.assertEqual(StringField().format(form), "text/plain")
            req.cnx.execute(
                'INSERT CWProperty X: X pkey "ui.default-text-format", X value "text/rest", X for_user U WHERE U login "admin"'
            )
            req.cnx.commit()
        with self.admin_access.web_request() as req:
            form = FieldsForm(req, None)
            self.assertEqual(StringField().format(form), "text/rest")

    def test_process_posted(self):
        class AForm(FieldsForm):
            anint = IntField()
            astring = StringField()

        with self.admin_access.web_request(
            anint="1", astring="2", _cw_fields="anint,astring"
        ) as req:
            form = AForm(req)
            self.assertEqual(form.process_posted(), {"anint": 1, "astring": "2"})
        with self.admin_access.web_request(
            anint="1a", astring="2b", _cw_fields="anint,astring"
        ) as req:
            form = AForm(req)
            self.assertRaises(ValidationError, form.process_posted)

    def test_jqdt_process_data(self):
        widget = JQueryDateTimePicker()
        field = DateTimeField("jqdt")
        with self.admin_access.web_request(
            **{"jqdt-date": "", "jqdt-time": "00:00"}
        ) as req:
            self._cw = req
            self.formvalues = {}
            date = widget.process_field_data(self, field)
            self.assertIsNone(date)


class EntityFieldsFormCWTC(WebCWTC):
    def test_form_field_choices(self):
        with self.admin_access.web_request() as req:
            b = req.create_entity(
                "BlogEntry", title="di mascii code", content="a best-seller"
            )
            t = req.create_entity("Tag", name="x")
            form1 = self.vreg["forms"].select("edition", req, entity=t)
            choices = [
                reid
                for rview, reid in form1.field_by_name(
                    "tags", "subject", t.e_schema
                ).choices(form1)
            ]
            self.assertIn(str(b.eid), choices)
            form2 = self.vreg["forms"].select("edition", req, entity=b)
            choices = [
                reid
                for rview, reid in form2.field_by_name(
                    "tags", "object", t.e_schema
                ).choices(form2)
            ]
            self.assertIn(str(t.eid), choices)

            b.cw_clear_all_caches()
            t.cw_clear_all_caches()
            req.cnx.execute("SET X tags Y WHERE X is Tag, Y is BlogEntry")

            choices = [
                reid
                for rview, reid in form1.field_by_name(
                    "tags", "subject", t.e_schema
                ).choices(form1)
            ]
            self.assertIn(str(b.eid), choices)
            choices = [
                reid
                for rview, reid in form2.field_by_name(
                    "tags", "object", t.e_schema
                ).choices(form2)
            ]
            self.assertIn(str(t.eid), choices)

    def test_form_field_choices_new_entity(self):
        with self.admin_access.web_request() as req:
            e = self.vreg["etypes"].etype_class("CWUser")(req)
            form = self.vreg["forms"].select("edition", req, entity=e)
            unrelated = [
                rview
                for rview, reid in form.field_by_name("in_group", "subject").choices(
                    form
                )
            ]
            # should be default groups but owners, i.e. managers, users, guests
            self.assertEqual(unrelated, ["guests", "managers", "users"])

    def test_consider_req_form_params(self):
        with self.admin_access.web_request() as req:
            e = self.vreg["etypes"].etype_class("CWUser")(req)
            e.eid = "A"
            with self.admin_access.web_request(login="toto") as toto_req:
                form = EntityFieldsForm(toto_req, None, entity=e)
                field = StringField(name="login", role="subject", eidparam=True)
                form.append_field(field)
                form.build_context({})
                self.assertEqual(field.widget.values(form, field), ("toto",))

    def test_linkto_field_duplication_inout(self):
        with self.admin_access.web_request() as req:
            e = self.vreg["etypes"].etype_class("CWUser")(req)
            e.eid = "A"
            e._cw = req
            geid = req.cnx.execute('CWGroup X WHERE X name "users"')[0][0]
            req.form["__linkto"] = "in_group:%s:subject" % geid
            form = self.vreg["forms"].select("edition", req, entity=e)
            form.content_type = "text/html"
            data = []
            form.render(w=data.append)
            pageinfo = self._check_html("\n".join(data), form, template=None)
            pageinfo.find_tag("select", False)
            ok = False
            for selectnode in pageinfo.matching_nodes(
                "select", name="from_in_group-subject:A"
            ):
                for optionnode in selectnode:
                    self.assertEqual(optionnode.get("value"), str(geid))
                    self.assertEqual(ok, False)
                    ok = True
            pageinfo.find_tag("input", False)
            self.assertFalse(list(pageinfo.matching_nodes("input", name="__linkto")))

    def test_reledit_composite_field(self):
        with self.admin_access.web_request() as req:
            rset = req.execute(
                'INSERT BlogEntry X: X title "cubicweb.org", X content "hop"'
            )
            form = self.vreg["views"].select(
                "reledit", req, rset=rset, row=0, rtype="content"
            )
            data = form.render(
                row=0, rtype="content", formid="base", action="edit_rtype"
            )
            self.assertIn("content_format", data)

    def test_form_generation_time(self):
        with self.admin_access.web_request() as req:
            e = req.create_entity("BlogEntry", title="cubicweb.org", content="hop")
            expected_field_name = "__form_generation_time:%d" % e.eid

            ts_before = time.time()
            form = self.vreg["forms"].select("edition", req, entity=e)
            ts_after = time.time()

            data = []
            form.render(action="edit", w=data.append)
            html_form = html.fromstring("".join(data)).forms[0]
            fields = dict(html_form.form_values())
            self.assertIn(expected_field_name, fields)
            ts = float(fields[expected_field_name])
            self.assertTrue(ts_before < ts < ts_after)

    # form tests ##############################################################

    def test_form_inheritance(self):
        with self.admin_access.web_request() as req:

            class CustomChangeStateForm(ChangeStateForm):
                hello = IntField(name="youlou")
                creation_date = DateTimeField(widget=DateTimePicker)

            form = CustomChangeStateForm(
                req, redirect_path="perdu.com", entity=req.user
            )
            data = []
            form.render(
                w=data.append,
                formvalues=dict(state=123, trcomment="", trcomment_format="text/plain"),
            )

    def test_change_state_form(self):
        with self.admin_access.web_request() as req:
            form = ChangeStateForm(req, redirect_path="perdu.com", entity=req.user)
            data = []
            form.render(
                w=data.append,
                formvalues=dict(state=123, trcomment="", trcomment_format="text/plain"),
            )

    # fields tests ############################################################

    def _render_entity_field(self, req, name, form):
        form.build_context({})
        renderer = FormRenderer(req)
        return form.field_by_name(name, "subject").render(form, renderer)

    def _test_richtextfield(self, req, expected):
        class RTFForm(EntityFieldsForm):
            description = RichTextField(eidparam=True, role="subject")

        state = self.vreg["etypes"].etype_class("State")(req)
        state.eid = "S"
        form = RTFForm(req, redirect_path="perdu.com", entity=state)
        # make it think it can use fck editor anyway
        form.field_by_name(
            "description", "subject"
        ).format = lambda form, field=None: "text/html"
        self.assertMultiLineEqual(
            self._render_entity_field(req, "description", form),
            expected % {"eid": state.eid},
        )

    def test_richtextfield_1(self):
        with self.admin_access.web_request() as req:
            req.use_fckeditor = lambda: False
            self._test_richtextfield(
                req,
                """<select id="description_format-subject:%(eid)s" name="description_format-subject:%(eid)s" size="1" style="display: block">
"""
                + """<option selected="selected" value="text/html">text/html</option>
<option value="text/markdown">text/markdown</option>
<option value="text/plain">text/plain</option>
<option value="text/rest">text/rest</option>
</select><textarea cols="80" id="description-subject:%(eid)s" name="description-subject:%(eid)s" onkeyup="autogrow(this)" rows="2"></textarea>""",
            )

    def test_richtextfield_2(self):
        with self.admin_access.web_request() as req:
            req.use_fckeditor = lambda: True
            self._test_richtextfield(
                req,
                '<input name="description_format-subject:%(eid)s" type="hidden" value="text/html" /><textarea cols="80" cubicweb:type="wysiwyg" id="description-subject:%(eid)s" name="description-subject:%(eid)s" onkeyup="autogrow(this)" rows="2"></textarea>',
            )

    def test_filefield(self):
        class FFForm(EntityFieldsForm):
            data = FileField(
                format_field=StringField(
                    name="data_format", max_length=50, eidparam=True, role="subject"
                ),
                encoding_field=StringField(
                    name="data_encoding", max_length=20, eidparam=True, role="subject"
                ),
                eidparam=True,
                role="subject",
            )

        with self.admin_access.web_request() as req:
            file = req.create_entity(
                "File",
                data_name="pouet.txt",
                data_encoding="UTF-8",
                data=Binary(b"new widgets system"),
            )
            form = FFForm(req, redirect_path="perdu.com", entity=file)
            self.assertMultiLineEqual(
                self._render_entity_field(req, "data", form),
                f"""<input id="data-subject:%(eid)s" name="data-subject:%(eid)s" type="file" value="" />
<a href="javascript: toggleVisibility(&#39;data-subject:%(eid)s-advanced&#39;)" title="show advanced fields"><img src="{BASE_URL}data/puce_down.png" alt="show advanced fields"/></a>
<div id="data-subject:%(eid)s-advanced" class="hidden">
<label for="data_format-subject:%(eid)s">data_format</label><input id="data_format-subject:%(eid)s" maxlength="50" name="data_format-subject:%(eid)s" size="45" type="text" value="text/plain" /><br/>
<label for="data_encoding-subject:%(eid)s">data_encoding</label><input id="data_encoding-subject:%(eid)s" maxlength="20" name="data_encoding-subject:%(eid)s" size="20" type="text" value="UTF-8" /><br/>
</div>
<br/>
<input name="data-subject__detach:%(eid)s" type="checkbox" />
detach attached file"""
                % {"eid": file.eid},
            )

    def test_editablefilefield(self):
        class EFFForm(EntityFieldsForm):
            data = EditableFileField(
                format_field=StringField(
                    "data_format", max_length=50, eidparam=True, role="subject"
                ),
                encoding_field=StringField(
                    "data_encoding", max_length=20, eidparam=True, role="subject"
                ),
                eidparam=True,
                role="subject",
            )

        with self.admin_access.web_request() as req:
            file = req.create_entity(
                "File",
                data_name="pouet.txt",
                data_encoding="UTF-8",
                data=Binary(b"new widgets system"),
            )
            form = EFFForm(req, redirect_path="perdu.com", entity=file)
            self.assertMultiLineEqual(
                self._render_entity_field(req, "data", form),
                f"""<input id="data-subject:%(eid)s" name="data-subject:%(eid)s" type="file" value="" />
<a href="javascript: toggleVisibility(&#39;data-subject:%(eid)s-advanced&#39;)" title="show advanced fields"><img src="{BASE_URL}data/puce_down.png" alt="show advanced fields"/></a>
<div id="data-subject:%(eid)s-advanced" class="hidden">
<label for="data_format-subject:%(eid)s">data_format</label><input id="data_format-subject:%(eid)s" maxlength="50" name="data_format-subject:%(eid)s" size="45" type="text" value="text/plain" /><br/>
<label for="data_encoding-subject:%(eid)s">data_encoding</label><input id="data_encoding-subject:%(eid)s" maxlength="20" name="data_encoding-subject:%(eid)s" size="20" type="text" value="UTF-8" /><br/>
</div>
<br/>
<input name="data-subject__detach:%(eid)s" type="checkbox" />
detach attached file
<p><b>You can either submit a new file using the browse button above, or choose to remove already uploaded file by checking the "detach attached file" check-box, or edit file content online with the widget below.</b></p>
<textarea cols="80" name="data-subject:%(eid)s" onkeyup="autogrow(this)" rows="3">new widgets system</textarea>"""
                % {"eid": file.eid},
            )

    def _modified_tzdatenaiss(self, eid, date_and_time_str=None):
        ctx = {}
        if date_and_time_str:
            datestr, timestr = date_and_time_str
            ctx["tzdatenaiss-subjectdate:%d" % eid] = datestr
            ctx["tzdatenaiss-subjecttime:%d" % eid] = timestr
        with self.admin_access.web_request(**ctx) as req:
            form = EntityFieldsForm(req, None, entity=req.entity_from_eid(eid))
            field = TZDatetimeField(name="tzdatenaiss", eidparam=True, role="subject")
            form.append_field(field)
            form.build_context({})
            return field.has_been_modified(form)

    def test_tzdatetimefield(self):
        """Comparison of the tz-aware database-stored value and the posted data
        should not crash, and the posted data should be considered UTC"""
        tzd = datetime.now(pytz.utc).replace(second=0, microsecond=0)
        datestr, timestr = tzd.strftime("%Y/%m/%d %H:%M").split()
        with self.admin_access.web_request() as req:
            eid = req.create_entity("Personne", nom="Flo", tzdatenaiss=tzd).eid
            req.cnx.commit()

        modified = self._modified_tzdatenaiss(eid, (datestr, timestr))
        self.assertFalse(modified)

        modified = self._modified_tzdatenaiss(eid, ("2016/05/04", "15:07"))
        self.assertTrue(modified)

        modified = self._modified_tzdatenaiss(eid, None)
        self.assertTrue(modified)

    def test_passwordfield(self):
        class PFForm(EntityFieldsForm):
            upassword = PasswordField(eidparam=True, role="subject")

        with self.admin_access.web_request() as req:
            form = PFForm(req, redirect_path="perdu.com", entity=req.user)
            self.assertMultiLineEqual(
                self._render_entity_field(req, "upassword", form),
                """<input id="upassword-subject:%(eid)s" name="upassword-subject:%(eid)s" type="password" value="" />
<br/>
<input name="upassword-subject-confirm:%(eid)s" type="password" value="" />
&#160;
<span class="emphasis">confirm password</span>"""
                % {"eid": req.user.eid},
            )

    # def test_datefield(self):
    #     class DFForm(EntityFieldsForm):
    #         creation_date = DateTimeField(widget=Input)
    #     form = DFForm(self.req, entity=self.entity)
    #     init, cur = (fromstring(self._render_entity_field(attr, form)).get('value')
    #                  for attr in ('edits-creation_date', 'creation_date'))
    #     self.assertEqual(init, cur)


if __name__ == "__main__":
    import unittest

    unittest.main()



# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""cubicweb.web.views.navigation unit tests"""

from logilab.common.testlib import unittest_main

from cubicweb_web.devtools.testlib import WebCWTC
from cubicweb_web.views.ibreadcrumbs import BreadCrumbEntityVComponent
from cubicweb_web.views.navigation import (
    PageNavigation,
    SortedNavigation,
    PageNavigationSelect,
)

BreadCrumbEntityVComponent.visible = True


class NavigationCWTC(WebCWTC):
    def test_navigation_selection_whatever(self):
        with self.admin_access.web_request() as req:
            rset = req.execute("Any X,N WHERE X name N")
            navcomp = self.vreg["components"].select("navigation", req, rset=rset)
            self.assertIsInstance(navcomp, PageNavigation)
            req.set_search_state("W:X:Y:Z")
            navcomp = self.vreg["components"].select("navigation", req, rset=rset)
            self.assertIsInstance(navcomp, PageNavigation)
            req.set_search_state("normal")

    def test_navigation_selection_ordered(self):
        with self.admin_access.web_request() as req:
            rset = req.execute("Any X,N ORDERBY N LIMIT 40 WHERE X name N")
            navcomp = self.vreg["components"].select(
                "navigation", req, rset=rset, page_size=20
            )
            self.assertIsInstance(navcomp, SortedNavigation)
            req.set_search_state("W:X:Y:Z")
            navcomp = self.vreg["components"].select(
                "navigation", req, rset=rset, page_size=20
            )
            self.assertIsInstance(navcomp, SortedNavigation)
            req.set_search_state("normal")
            navcomp.render()

    def test_navigation_selection_large_rset(self):
        with self.admin_access.web_request() as req:
            rset = req.execute("Any X,N LIMIT 120 WHERE X name N")
            navcomp = self.vreg["components"].select(
                "navigation", req, rset=rset, page_size=20
            )
            self.assertIsInstance(navcomp, PageNavigationSelect)
            rset = req.execute("Any X,N ORDERBY N LIMIT 120 WHERE X name N")
            navcomp = self.vreg["components"].select(
                "navigation", req, rset=rset, page_size=20
            )
            self.assertIsInstance(navcomp, PageNavigationSelect)

    def test_navigation_selection_not_enough_1(self):
        with self.admin_access.web_request() as req:
            rset = req.execute("Any X,N LIMIT 10 WHERE X name N")
            navcomp = self.vreg["components"].select_or_none(
                "navigation", req, rset=rset
            )
            self.assertEqual(navcomp, None)
            req.set_search_state("W:X:Y:Z")
            navcomp = self.vreg["components"].select_or_none(
                "navigation", req, rset=rset
            )
            self.assertEqual(navcomp, None)
            req.set_search_state("normal")

    def test_navigation_selection_not_enough_2(self):
        with self.admin_access.web_request() as req:
            rset = req.execute(
                "Any N, COUNT(RDEF) GROUPBY N ORDERBY N "
                "WHERE RDEF relation_type RT, RT name N"
            )
            navcomp = self.vreg["components"].select("navigation", req, rset=rset)
            self.assertIsInstance(navcomp, SortedNavigation)
            req.set_search_state("W:X:Y:Z")
            navcomp = self.vreg["components"].select("navigation", req, rset=rset)
            self.assertIsInstance(navcomp, SortedNavigation)

    def test_navigation_selection_wrong_boundary(self):
        with self.admin_access.web_request() as req:
            rset = req.execute("Any X,N WHERE X name N")
            req.form["__start"] = 1000000
            navcomp = self.vreg["components"].select("navigation", req, rset=rset)
            navcomp.render()

    def test_sorted_navigation_1(self):
        with self.admin_access.web_request() as req:
            rset = req.execute("Any RDEF ORDERBY RT WHERE RDEF relation_type RT")
            navcomp = self.vreg["components"].select("navigation", req, rset=rset)
            navcomp.render()

    def test_sorted_navigation_2(self):
        with self.admin_access.web_request() as req:
            rset = req.execute("Any RDEF ORDERBY RDEF WHERE RDEF relation_type RT")
            navcomp = self.vreg["components"].select("navigation", req, rset=rset)
            navcomp.render()

    def test_sorted_navigation_3(self):
        with self.admin_access.web_request() as req:
            rset = req.execute("CWAttribute RDEF ORDERBY RDEF")
            navcomp = self.vreg["components"].select("navigation", req, rset=rset)
            navcomp.render()

    def test_sorted_navigation_4(self):
        with self.admin_access.web_request() as req:
            rset = req.execute(
                "Any RDEF ORDERBY N " "WHERE RDEF relation_type RT, RT name N"
            )
            navcomp = self.vreg["components"].select("navigation", req, rset=rset)
            navcomp.render()

    def test_sorted_navigation_5(self):
        with self.admin_access.web_request() as req:
            rset = req.execute(
                "Any N, COUNT(RDEF) GROUPBY N ORDERBY N "
                "WHERE RDEF relation_type RT, RT name N"
            )
            navcomp = self.vreg["components"].select("navigation", rset.req, rset=rset)
            navcomp.render()


if __name__ == "__main__":
    unittest_main()



# copyright 2003-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""contains utility functions and some visual component to restrict results of
a search

"""


from logilab.common.deprecation import callable_moved

insert_attr_select_relation = callable_moved(
    "cubicweb_web.facet", "insert_attr_select_relation"
)



from cubicweb_web import component
from cubicweb_web.views import ajaxcontroller


@ajaxcontroller.ajaxfunc
def tag_entity(self, eid, taglist):
    execute = self._cw.execute
    # get list of tag for this entity
    tagged_by = {
        tagname
        for (tagname,) in execute(
            "Any N WHERE T name N, T tags X, X eid %(x)s", {"x": eid}
        )
    }
    for tagname in taglist:
        tagname = tagname.strip()
        if not tagname or tagname in tagged_by:
            continue
        tagrset = execute("Tag T WHERE T name %(name)s", {"name": tagname})
        if tagrset:
            rql = "SET T tags X WHERE T eid %(t)s, X eid %(x)s"
            execute(rql, {"t": tagrset[0][0], "x": eid})
        else:
            rql = "INSERT Tag T: T name %(name)s, T tags X WHERE X eid %(x)s"
            execute(rql, {"name": tagname, "x": eid})


class TagsBox(component.AjaxEditRelationCtxComponent):
    __regid__ = "tags_box"
    rtype = "tags"
    role = "object"
